#import <string.h>
#import <Cocoa/Cocoa.h>
#import <IOKit/IOMessage.h>
#import <IOKit/pwr_mgt/IOPMLib.h>
#import "MVChatConnection.h"
#import "MVChatWindowController.h"
#import "MVChatPluginManager.h"
#import "NSAttributedStringAdditions.h"
#import "firetalk.h"

typedef void (*firetalk_callback)(firetalk_t, void *, ...);
typedef void (*firetalk_subcode_callback)(firetalk_t, void *, const char * const, const char * const, const char * const);

#pragma mark -

NSString *MVChatConnectionWillConnectNotification = @"MVChatConnectionWillConnectNotification";
NSString *MVChatConnectionDidConnectNotification = @"MVChatConnectionDidConnectNotification";
NSString *MVChatConnectionDidNotConnectNotification = @"MVChatConnectionDidNotConnectNotification";
NSString *MVChatConnectionWillDisconnectNotification = @"MVChatConnectionWillDisconnectNotification";
NSString *MVChatConnectionDidDisconnectNotification = @"MVChatConnectionDidDisconnectNotification";
NSString *MVChatConnectionDidGetErrorNotification = @"MVChatConnectionDidGetErrorNotification";

NSString *MVChatConnectionNeedPasswordNotification = @"MVChatConnectionNeedPasswordNotification";
NSString *MVChatConnectionGotPrivateMessageNotification = @"MVChatConnectionGotPrivateMessageNotification";

NSString *MVChatConnectionBuddyIsOnlineNotification = @"MVChatConnectionBuddyIsOnlineNotification";
NSString *MVChatConnectionBuddyIsOfflineNotification = @"MVChatConnectionBuddyIsOfflineNotification";
NSString *MVChatConnectionBuddyIsAwayNotification = @"MVChatConnectionBuddyIsAwayNotification";
NSString *MVChatConnectionBuddyIsUnawayNotification = @"MVChatConnectionBuddyIsUnawayNotification";
NSString *MVChatConnectionBuddyIsIdleNotification = @"MVChatConnectionBuddyIsIdleNotification";

NSString *MVChatConnectionGotUserInfoNotification = @"MVChatConnectionGotUserInfoNotification";
NSString *MVChatConnectionGotRoomInfoNotification = @"MVChatConnectionGotRoomInfoNotification";

NSString *MVChatConnectionJoinedRoomNotification = @"MVChatConnectionJoinedRoomNotification";
NSString *MVChatConnectionLeftRoomNotification = @"MVChatConnectionLeftRoomNotification";
NSString *MVChatConnectionUserJoinedRoomNotification = @"MVChatConnectionUserJoinedRoomNotification";
NSString *MVChatConnectionUserLeftRoomNotification = @"MVChatConnectionUserLeftRoomNotification";
NSString *MVChatConnectionUserNicknameChangedNotification = @"MVChatConnectionUserNicknameChangedNotification";
NSString *MVChatConnectionUserOppedInRoomNotification = @"MVChatConnectionUserOppedInRoomNotification";
NSString *MVChatConnectionUserDeoppedInRoomNotification = @"MVChatConnectionUserDeoppedInRoomNotification";
NSString *MVChatConnectionUserVoicedInRoomNotification = @"MVChatConnectionUserVoicedInRoomNotification";
NSString *MVChatConnectionUserDevoicedInRoomNotification = @"MVChatConnectionUserDevoicedInRoomNotification";
NSString *MVChatConnectionUserKickedFromRoomNotification = @"MVChatConnectionUserKickedFromRoomNotification";
NSString *MVChatConnectionGotRoomMessageNotification = @"MVChatConnectionGotRoomMessageNotification";
NSString *MVChatConnectionGotRoomTopicNotification = @"MVChatConnectionGotRoomTopicNotification";

NSString *MVChatConnectionKickedFromRoomNotification = @"MVChatConnectionKickedFromRoomNotification";
NSString *MVChatConnectionInvitedToRoomNotification = @"MVChatConnectionInvitedToRoomNotification";

NSString *MVChatConnectionNicknameAcceptedNotification = @"MVChatConnectionNicknameAcceptedNotification";
NSString *MVChatConnectionNicknameRejectedNotification = @"MVChatConnectionNicknameRejectedNotification";

NSString *MVChatConnectionFileTransferAvailableNotification = @"MVChatConnectionFileTransferAvailableNotification";
NSString *MVChatConnectionFileTransferOfferedNotification = @"MVChatConnectionFileTransferOfferedNotification";
NSString *MVChatConnectionFileTransferStartedNotification = @"MVChatConnectionFileTransferStartedNotification";
NSString *MVChatConnectionFileTransferFinishedNotification = @"MVChatConnectionFileTransferFinishedNotification";
NSString *MVChatConnectionFileTransferErrorNotification = @"MVChatConnectionFileTransferErrorNotification";
NSString *MVChatConnectionFileTransferStatusNotification = @"MVChatConnectionFileTransferStatusNotification";

NSString *MVChatConnectionSubcodeRequestNotification = @"MVChatConnectionSubcodeRequestNotification";
NSString *MVChatConnectionSubcodeReplyNotification = @"MVChatConnectionSubcodeReplyNotification";

@interface MVChatConnection (MVChatConnectionPrivate)
- (io_connect_t) _powerConnection;
- (firetalk_t) _firetalkConnection;
- (void) _executeRunLoopCheck:(NSTimer *) timer;
- (void) _registerCallbacks;
- (void) _registerForSleepNotifications;
- (void) _deregisterForSleepNotifications;
- (void) _addRoomToCache:(NSString *) room withUsers:(int) users andTopic:(NSString *) topic;
- (void) _setBacklogDelay:(NSTimeInterval) delay;
- (void) _setStatus:(MVChatConnectionStatus) status;
- (void) _willConnect;
- (void) _didConnect;
- (void) _didNotConnect;
- (void) _willDisconnect;
- (void) _didDisconnect;
@end

#pragma mark -

void MVChatHandlePowerChange( void *refcon, io_service_t service, natural_t messageType, void *messageArgument ) {
	MVChatConnection *self = refcon;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatHandlePowerChange()\n");
#endif
    
    switch( messageType ) {
		case kIOMessageSystemWillRestart:
		case kIOMessageSystemWillPowerOff:
		case kIOMessageSystemWillSleep:
		case kIOMessageDeviceWillPowerOff:
			if( [self isConnected] ) {
				[self disconnect];
				[self _setStatus:MVChatConnectionSuspendedStatus];
			}
			IOAllowPowerChange( [self _powerConnection], (long) messageArgument );
			break;
		case kIOMessageCanSystemPowerOff:
		case kIOMessageCanSystemSleep:
		case kIOMessageCanDevicePowerOff:
			IOAllowPowerChange( [self _powerConnection], (long) messageArgument );
			break;
		case kIOMessageSystemWillNotSleep:
		case kIOMessageSystemWillNotPowerOff:
		case kIOMessageSystemHasPoweredOn:
		case kIOMessageDeviceWillNotPowerOff:
		case kIOMessageDeviceHasPoweredOn:
			if( [self status] == MVChatConnectionSuspendedStatus ) [self connect];
			break;
	}
}

#pragma mark -

void MVChatLoad( void *c, void *cs, const char * const nickname ) {

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatLoad()\n");
#endif
}

void MVChatConnected( void *c, void *cs ) {
	MVChatConnection *self = cs;
    NSEnumerator *enumerator = [[MVChatWindowController roomChatWindowsForConnection:self] keyEnumerator];
	id key = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatConnected()\n");
#endif

	[self _didConnect];
	while( ( key = [enumerator nextObject] ) )
		[self joinChatForRoom:key];
}

void MVChatConnectionFailed( void *c, void *cs, const int error, const char * const reason ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatConnectionFailed()\n");
#endif

    [self _didNotConnect];
	switch( error ) {
		case FE_SOCKET:
		case FE_RESOLV:
			if( NSRunCriticalAlertPanel( NSLocalizedString( @"Could not connect to Chat server", "chat invalid password dialog title" ), NSLocalizedString( @"The server is disconnected or refusing connections from your computer. Make sure you are conencted to the internet and have access to the server.", "chat invalid password dialog message" ), NSLocalizedString( @"Retry", "retry connecting to server" ), @"Cancel", nil ) == NSOKButton )
				[self connect];
			break;
		case FE_BADUSERPASS:
			NSRunCriticalAlertPanel( NSLocalizedString( @"Your Chat password is invalid", "chat invalid password dialog title" ), NSLocalizedString( @"The password you specified is invalid or a connection could not be made without a proper password. Make sure you have access to the server.", "chat invalid password dialog message" ), nil, nil, nil );
			break;
		case FE_BADUSER:
			NSRunCriticalAlertPanel( NSLocalizedString( @"Your Chat nickname could not be used", "chat invalid nickname dialog title" ), [NSString stringWithFormat:NSLocalizedString( @"The nickname you specified is in use or invalid on this server. A connection could not be made with '%@' as your nickname.", "chat invalid nicknames dialog message" ), [self nickname]], nil, nil, nil );
			break;
		default:
			NSRunCriticalAlertPanel( NSLocalizedString( @"An error occured while connecting", "chat connecting error dialog title" ), [NSString stringWithFormat:NSLocalizedString( @"The connection could not be made. %s.", "unknown connection error dialog message" ), reason], nil, nil, nil );
			break;
	}
}

void MVChatDisconnect( void *c, void *cs, const int error ) {
	MVChatConnection *self = cs;
	MVChatConnectionStatus status = [self status];

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatDisconnect()\n");
#endif

    [self _didDisconnect];
	switch( error ) {
		case FE_USERDISCONNECT:
			break;
		case FE_DISCONNECT:
		case FE_PACKET:
		case FE_PACKETSIZE:
			if( status == MVChatConnectionConnectedStatus ) {
				if( NSRunCriticalAlertPanel( NSLocalizedString( @"You have been disconnected", "title of the you have been disconnected error" ), NSLocalizedString( @"The server may have shutdown for maintenance, or the connection was broken between your computer and the server. Check your connection and try again.", "connection dropped" ), NSLocalizedString( @"Reconnect", "reconnect to server button" ), @"Cancel", nil ) == NSOKButton )
					[self connect];
			} else {
				if( NSRunCriticalAlertPanel( NSLocalizedString( @"Could not connect", "title of the could not connect error" ), NSLocalizedString( @"The server may be down for maintenance, or the connection was broken between your computer and the server. Check your connection and try again.", "connection dropped" ), NSLocalizedString( @"Retry", "retry connecting to server" ), @"Cancel", nil ) == NSOKButton )
					[self connect];
			}
			break;
		default:
			NSRunCriticalAlertPanel( NSLocalizedString( @"You have been disconnected", "title of the you have been disconnected error" ), [NSString stringWithFormat:NSLocalizedString( @"The connection was terminated between your computer and the server. %s.", "unknown disconnection error dialog message" ), firetalk_strerror( error )], nil, nil, nil );
			break;
	}
}

void MVChatError( void *c, void *cs, const int error, const char * const roomoruser ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatError()\n");
#endif

    switch( error ) {
		case FE_BADUSER:
			if( roomoruser && strlen( roomoruser ) >= 1 && strchr( "#&+", roomoruser[0] ) )
				[[MVChatWindowController chatWindowForRoom:[[NSString stringWithUTF8String:roomoruser] lowercaseString] withConnection:self ifExists:YES] disconnected];
			else if( roomoruser ) [[MVChatWindowController chatWindowWithUser:[NSString stringWithUTF8String:roomoruser] withConnection:self ifExists:YES] unavailable];
			else {
				[MVChatWindowController changeSelfInChatWindowsTo:[self nickname] forConnection:self];
				NSRunCriticalAlertPanel( NSLocalizedString( @"Your Chat nickname could not be used", "chat invalid nickname dialog title" ), NSLocalizedString( @"The nickname you specified is in use or invalid on this server.", "chat invalid nickname dialog message" ), nil, nil, nil );
			}
			break;
		default:
			NSRunCriticalAlertPanel( NSLocalizedString( @"An error occured", "unknown error dialog title" ), [NSString stringWithFormat:NSLocalizedString( @"An error occured when dealing with %@. %s.", "unknown error dialog message" ), ( roomoruser ? [NSString stringWithUTF8String:roomoruser] : NSLocalizedString( @"server", "singular server label" ) ), firetalk_strerror( error )], nil, nil, nil );
			break;
	}
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionDidGetErrorNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:( roomoruser ? [NSString stringWithUTF8String:roomoruser] : [NSNull null] ), @"target", [NSNumber numberWithInt:error], @"error", nil]];
}

void MVChatBackLog( void *c, void *cs, const double backlog ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatBackLog()\n");
#endif

    [self _setBacklogDelay:backlog];
}

#pragma mark -

void MVChatNeedPassword( void *c, void *cs, char *password, const int size ) {
	MVChatConnection *self = cs;
    const char *pass = [[self password] UTF8String];

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatNeedPassword()\n");
#endif
    
    if( ! pass ) {
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionNeedPasswordNotification object:self userInfo:nil];
    } else {
        strncpy( password, pass, size );
    }
}

#pragma mark -

void MVChatGetMessage( void *c, void *cs, const char * const who, const int automessage, const char * const message ) {
	MVChatConnection *self = cs;
	MVChatWindowController *window = nil;
	NSData *msgData = nil;
	NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );
	NSCParameterAssert( message != NULL );

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatGetMessage()\n");
#endif

    msgData = [NSData dataWithBytes:message length:strlen( message )];

	window = [MVChatWindowController chatWindowWithUser:[NSString stringWithUTF8String:who] withConnection:self ifExists:YES];
	if( ! automessage || window ) {
		[[MVChatWindowController chatWindowWithUser:[NSString stringWithUTF8String:who] withConnection:self ifExists:NO] addHTMLMessageToDisplay:msgData fromUser:[NSString stringWithUTF8String:who] asAction:NO asAlert:NO];
	} else NSLog( @"%@ notice: %@", [NSString stringWithUTF8String:who], [NSString stringWithUTF8String:message] );
	[MVChatWindowController updateChatWindowsMember:[NSString stringWithUTF8String:who] withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotPrivateMessageNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"from", [NSNumber numberWithBool:automessage], @"auto", msgData, @"message", nil]];
}

void MVChatGetAction( void *c, void *cs, const char * const who, const int automessage, const char * const message ) {
	MVChatConnection *self = cs;
	NSData *msgData = nil;
	NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );
	NSCParameterAssert( message != NULL );

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatGetAction()\n");
#endif

    msgData = [NSData dataWithBytes:message length:strlen( message )];

	if( ! automessage ) {
		[[MVChatWindowController chatWindowWithUser:[NSString stringWithUTF8String:who] withConnection:self ifExists:NO] addHTMLMessageToDisplay:msgData fromUser:[NSString stringWithUTF8String:who] asAction:YES asAlert:NO];
	} else NSLog( @"%@ notice: %@", [NSString stringWithUTF8String:who], [NSString stringWithUTF8String:message] );
	[MVChatWindowController updateChatWindowsMember:[NSString stringWithUTF8String:who] withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotPrivateMessageNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"from", [NSNumber numberWithBool:automessage], @"auto", [NSNumber numberWithBool:YES], @"action", msgData, @"message", nil]];
}

#pragma mark -

void MVChatBuddyOnline( void *c, void *cs, const char * const who ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatBuddyOnline()\n");
#endif

	NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );
    [[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionBuddyIsOnlineNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"who", nil]];
}

void MVChatBuddyOffline( void *c, void *cs, const char * const who ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatBuddyOffline()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );

    [[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionBuddyIsOfflineNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"who", nil]];
}

void MVChatBuddyAway( void *c, void *cs, const char * const who ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatBuddyAway()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );

    [[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionBuddyIsAwayNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"who", nil]];
}

void MVChatBuddyUnaway( void *c, void *cs, const char * const who ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatBuddyUnaway()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );

    [[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionBuddyIsUnawayNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"who", nil]];
}

void MVChatBuddyGotIdle( void *c, void *cs, const char * const who, const long idletime ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatBuddyGotIdle()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );
    
	[MVChatWindowController updateChatWindowsMember:[NSString stringWithUTF8String:who] withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:idletime] forKey:@"idle"] forConnection:self];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionBuddyIsIdleNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"who", [NSNumber numberWithLong:idletime], @"idle", nil]];
}

void MVChatListBuddy( void *c, void *cs, const char * const nickname, const int online, const int away ) {

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( nickname != NULL );

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatListBuddy()\n");
#endif
}

#pragma mark -

void MVChatGotInfo( void *c, void *cs, const char * const who, const char * const info, const int warning, const int idle, const int flags ) {
	MVChatConnection *self = cs;
	NSDictionary *infoDic = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatGotInfo()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( who != NULL );

    infoDic = [NSDictionary dictionaryWithObjectsAndKeys:( info ? [NSString stringWithUTF8String:info] : [NSNull null] ), @"info", [NSNumber numberWithUnsignedInt:warning], @"warning", [NSNumber numberWithUnsignedInt:idle], @"idle", [NSNumber numberWithUnsignedInt:flags], @"flags", nil];
	[MVChatWindowController updateChatWindowsMember:[NSString stringWithUTF8String:who] withInfo:infoDic forConnection:self];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotUserInfoNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:who], @"who", infoDic, @"info", nil]];
}

void MVChatListRoom( void *c, void *cs, const char * const room, const int users, const char * const topic ) {
	MVChatConnection *self = cs;
	NSString *r = nil, *t = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatListRoom()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( topic != NULL );

    r = [NSString stringWithUTF8String:room];
	t = [NSString stringWithUTF8String:topic];
	[self _addRoomToCache:r withUsers:users andTopic:t];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotRoomInfoNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:r, @"room", [NSNumber numberWithUnsignedInt:users], @"users", t, @"topic", nil]];
}

#pragma mark -

void MVChatJoinedRoom( void *c, void *cs, const char * const room ) {
	MVChatConnection *self = cs;
	MVChatWindowController *window = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatJoinedRoom()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );

    window = [MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:NO];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionJoinedRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", nil]];
}

void MVChatLeftRoom( void *c, void *cs, const char * const room ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatLeftRoom()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );

    [MVChatWindowController disposeWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionLeftRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", nil]];
}

void MVChatGetRoomMessage( void *c, void *cs, const char * const room, const char * const from, const int automessage, const char * message ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatGetRoomMessage()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( from != NULL );
	NSCParameterAssert( message != NULL );

    {
		MVChatWindowController *window = [MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES];
		NSData *msgData = [NSData dataWithBytes:message length:strlen( message )];
		[window addHTMLMessageToDisplay:msgData fromUser:[NSString stringWithUTF8String:from] asAction:NO asAlert:NO];

		[MVChatWindowController updateChatWindowsMember:[NSString stringWithUTF8String:from] withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotRoomMessageNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:from], @"from", [NSNumber numberWithBool:automessage], @"auto", msgData, @"message", nil]];
	}
}

void MVChatGetRoomAction( void *c, void *cs, const char * const room, const char * const from, const int automessage, const char * message ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatGetRoomAction()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( from != NULL );
	NSCParameterAssert( message != NULL );
	{
		MVChatWindowController *window = [MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES];
		NSData *msgData = [NSData dataWithBytes:message length:strlen(message)];
		[window addHTMLMessageToDisplay:msgData fromUser:[NSString stringWithUTF8String:from] asAction:YES asAlert:NO];

		[MVChatWindowController updateChatWindowsMember:[NSString stringWithUTF8String:from] withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotRoomMessageNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:from], @"from", [NSNumber numberWithBool:automessage], @"auto", [NSNumber numberWithBool:YES], @"action", msgData, @"message", nil]];
	}
}

#pragma mark -

void MVChatKicked( void *c, void *cs, const char * const room, const char * const by, const char * const reason ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatKicked()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	{
		NSData *msgData = [NSData dataWithBytes:reason length:(reason ? strlen(reason) : 0)];
		[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] kickedFromChatBy:(by?[NSString stringWithUTF8String:by]:nil) forReason:( reason ? msgData : nil )];
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionKickedFromRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", ( by ? [NSString stringWithUTF8String:by] : [NSNull null] ), @"who", ( reason ? (id) msgData : (id) [NSNull null] ), @"reason", nil]];
	}
}

void MVChatInvited( void *c, void *cs, const char * const room, const char * const from, const char * message ) {
	NSWindow *win = nil;
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatInvited()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( from != NULL );

    win = [[MVChatWindowController chatWindowWithUser:[NSString stringWithUTF8String:from] withConnection:self ifExists:YES] window];
	if( NSRunAlertPanelRelativeToWindow( NSLocalizedString( @"Invited to Chat", "invited to a chat room - sheet title" ), [NSString stringWithFormat:NSLocalizedString( @"You have been invited to chat in the %@ room by %@.", "invited to chat room description - sheet message" ), [NSString stringWithUTF8String:room], [NSString stringWithUTF8String:from]], nil, NSLocalizedString( @"Refuse", "refuse button name" ), nil, win ) == NSOKButton ) {
		[self joinChatForRoom:[NSString stringWithUTF8String:room]];
	}
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionInvitedToRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:from], @"from", nil]];
}

void MVChatGotTopic( void *c, void *cs, const char * const room, const char * const topic, const char * const author ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatGotTopic()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	{
		NSData *msgData = [NSData dataWithBytes:topic length:(topic ? strlen(topic) : 0)];
		[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] changeTopic:( topic ? msgData : nil ) by:( author ? [NSString stringWithUTF8String:author] : nil )];
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionGotRoomTopicNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", ( author ? (id) [NSString stringWithUTF8String:author] : (id) [NSNull null] ), @"author", ( topic ? (id) msgData : (id) [NSNull null] ), @"topic", nil]];
	}
}

#pragma mark -

void MVChatUserJoinedRoom( void *c, void *cs, const char * const room, const char * const who, const int previousmember ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserJoinedRoom()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	firetalk_im_get_info( c, who, 0 );
	[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] addMemberToChat:[NSString stringWithUTF8String:who] asPreviousMember:(BOOL) previousmember];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserJoinedRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", [NSNumber numberWithBool:previousmember], @"previousMember", nil]];
}

void MVChatUserLeftRoom( void *c, void *cs, const char * const room, const char * const who, const char * const reason ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserLeftRoom()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	{
		NSData *msgData = [NSData dataWithBytes:reason length:(reason ? strlen(reason) : 0)];
		[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] removeChatMember:[NSString stringWithUTF8String:who] withReason:( reason ? msgData : nil)];
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserLeftRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", ( reason ? (id) msgData : (id) [NSNull null] ), @"reason", nil]];
	}
}

void MVChatUserNicknameChanged( void *c, void *cs, const char * const room, const char * const oldnick, const char * const newnick ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserNicknameChanged()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( oldnick != NULL );
	NSCParameterAssert( newnick != NULL );

	[MVChatWindowController changeMemberInChatWindowsFrom:[NSString stringWithUTF8String:oldnick] to:[NSString stringWithUTF8String:newnick] forConnection:self];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserNicknameChangedNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:oldnick], @"oldNickname", [NSString stringWithUTF8String:newnick], @"newNickname", nil]];
}

#pragma mark -

void MVChatNewNickname( void *c, void *cs, const char * const newnick ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatNewNickname()\n");
#endif

//Markus - unused    NSCParameterAssert(c);
	NSCParameterAssert(newnick);
    
	[MVChatWindowController changeSelfInChatWindowsTo:[NSString stringWithUTF8String:newnick] forConnection:self];
    [self setNickname:[NSString stringWithUTF8String:newnick]];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionNicknameAcceptedNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:newnick], @"nickname", nil]];
}

#pragma mark -

void MVChatUserOpped( void *c, void *cs, const char * const room, const char * const who, const char * const by ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserOpped()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] promoteChatMember:[NSString stringWithUTF8String:who] by:(by?[NSString stringWithUTF8String:by]:nil)];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserOppedInRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", (by?[NSString stringWithUTF8String:by]:[NSNull null]), @"by", nil]];
}

void MVChatUserDeopped( void *c, void *cs, const char * const room, const char * const who, const char * const by ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserDeopped()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] demoteChatMember:[NSString stringWithUTF8String:who] by:(by?[NSString stringWithUTF8String:by]:nil)];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserDeoppedInRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", (by?[NSString stringWithUTF8String:by]:[NSNull null]), @"by", nil]];
}

void MVChatUserVoiced( void *c, void *cs, const char * const room, const char * const who, const char * const by ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserVoiced()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] voiceChatMember:[NSString stringWithUTF8String:who] by:(by?[NSString stringWithUTF8String:by]:nil)];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserVoicedInRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", (by?[NSString stringWithUTF8String:by]:[NSNull null]), @"by", nil]];
}

void MVChatUserDevoiced( void *c, void *cs, const char * const room, const char * const who, const char * const by ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserDevoiced()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] devoiceChatMember:[NSString stringWithUTF8String:who] by:(by?[NSString stringWithUTF8String:by]:nil)];
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserDevoicedInRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", (by?[NSString stringWithUTF8String:by]:[NSNull null]), @"by", nil]];
}

void MVChatUserKicked( void *c, void *cs, const char * const room, const char * const who, const char * const by, const char * const reason ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatUserKicked()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( room != NULL );
	NSCParameterAssert( who != NULL );
	{
		NSData *msgData = [NSData dataWithBytes:reason length:(reason ? strlen(reason) : 0)];
		[[MVChatWindowController chatWindowForRoom:[NSString stringWithUTF8String:room] withConnection:self ifExists:YES] chatMemberKicked:[NSString stringWithUTF8String:who] by:(by?[NSString stringWithUTF8String:by]:nil) forReason:( reason ? msgData : nil )];
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionUserKickedFromRoomNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:room], @"room", [NSString stringWithUTF8String:who], @"who", (by?[NSString stringWithUTF8String:by]:[NSNull null]), @"by", msgData, @"reason", nil]];
	}
}

#pragma mark -

void MVChatFileTransferAccept( void *c, void *cs, const void * const filehandle, const char * const from, const char * const filename, const long size ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatFileTransferAccept()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( filehandle != NULL );
	NSCParameterAssert( from != NULL );
	NSCParameterAssert( filename != NULL );
	NSCParameterAssert( size >= 0 );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionFileTransferAvailableNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%x", filehandle], @"identifier", [NSString stringWithUTF8String:from], @"from", [NSString stringWithUTF8String:filename], @"filename", [NSNumber numberWithUnsignedLong:size], @"size", nil]];
}

void MVChatFileTransferStart( void *c, void *cs, const void * const filehandle, const void * const clientfilestruct ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatFileTransferStart()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( filehandle != NULL );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionFileTransferStartedNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%x", filehandle], @"identifier", nil]];
}

void MVChatFileTransferFinish( void *c, void *cs, const void * const filehandle, const void * const clientfilestruct, const long size ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatFileTransferFinish()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( filehandle != NULL );
	NSCParameterAssert( size >= 0 );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionFileTransferFinishedNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%x", filehandle], @"identifier", [NSNumber numberWithUnsignedLong:size], @"size", nil]];
}

void MVChatFileTransferError( void *c, void *cs, const void * const filehandle, const void * const clientfilestruct, const int error ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatFileTransferError()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( filehandle != NULL );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionFileTransferErrorNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%x", filehandle], @"identifier", [NSNumber numberWithInt:error], @"error", nil]];
}

void MVChatFileTransferStatus( void *c, void *cs, const void * const filehandle, const void * const clientfilestruct, const long bytes, const long size ) {
	MVChatConnection *self = cs;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatFileTransferStatus()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( filehandle != NULL );
	NSCParameterAssert( bytes >= 0 );
	NSCParameterAssert( size >= 0 );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionFileTransferStatusNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%x", filehandle], @"identifier", [NSNumber numberWithUnsignedLong:bytes], @"transfered", [NSNumber numberWithUnsignedLong:size], @"size", nil]];
}

#pragma mark -

void MVChatSubcodeRequest( void *c, void *cs, const char * const from, const char * const command, const char * const args ) {
	MVChatConnection *self = cs;
	NSEnumerator *enumerator = nil;
	id item = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatSubcodeRequest()\n");
#endif

    NSCParameterAssert( c != NULL );
	NSCParameterAssert( from != NULL );
	NSCParameterAssert( command != NULL );

	enumerator = [[[MVChatPluginManager defaultManager] pluginsThatRespondToSelector:@selector( processSubcodeRequest:withArguments:fromUser:forConnection: )] objectEnumerator];
	while( ( item = [enumerator nextObject] ) )
		if( [item processSubcodeRequest:[NSString stringWithUTF8String:command] withArguments:[NSString stringWithUTF8String:args] fromUser:[NSString stringWithUTF8String:from] forConnection:self] ) return;

	if( ! strcasecmp( command, "VERSION" ) ) {
		NSDictionary *systemVersion = [NSDictionary dictionaryWithContentsOfFile:@"/System/Library/CoreServices/SystemVersion.plist"];
		NSDictionary *clientVersion = [[NSBundle mainBundle] infoDictionary];
		NSString *reply = [NSString stringWithFormat:@"%@ %@ - %@ %@ - http://www.javelin.cc?colloquy", [clientVersion objectForKey:@"CFBundleName"], [clientVersion objectForKey:@"CFBundleShortVersionString"], [systemVersion objectForKey:@"ProductName"], [systemVersion objectForKey:@"ProductUserVisibleVersion"]];
		[self sendSubcodeReply:@"VERSION" toUser:[NSString stringWithUTF8String:from] withArguments:reply];
		return;
	} else 	if( ! strcasecmp( command, "USERINFO" ) ) {
		[self sendSubcodeReply:@"USERINFO" toUser:[NSString stringWithUTF8String:from] withArguments:@"..."];
		return;
	} else 	if( ! strcasecmp( command, "URL" ) ) {
		[self sendSubcodeReply:@"URL" toUser:[NSString stringWithUTF8String:from] withArguments:@"..."];
		return;
	}

	NSLog( @"subcode request from '%s' for '%s' with args '%s'", from, command, args );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionSubcodeRequestNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:from], @"from", [NSString stringWithUTF8String:command], @"command", (args ? (id) [NSString stringWithUTF8String:args] : (id) [NSNull null]), @"arguments", nil]];
}

void MVChatSubcodeReply( void *c, void *cs, const char * const from, const char * const command, const char * const args ) {
	MVChatConnection *self = cs;
	NSEnumerator *enumerator = nil;
	id item = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("MVChatSubcodeReply()\n");
#endif
    
	NSCParameterAssert( c != NULL );
	NSCParameterAssert( from != NULL );
	NSCParameterAssert( command != NULL );

	enumerator = [[[MVChatPluginManager defaultManager] pluginsThatRespondToSelector:@selector( processSubcodeReply:withArguments:fromUser:forConnection: )] objectEnumerator];
	while( ( item = [enumerator nextObject] ) )
		if( [item processSubcodeReply:[NSString stringWithUTF8String:command] withArguments:[NSString stringWithUTF8String:args] fromUser:[NSString stringWithUTF8String:from] forConnection:self] ) return;

	NSLog( @"subcode reply from '%s' for '%s' with args '%s'", from, command, args );
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionSubcodeReplyNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithUTF8String:from], @"from", [NSString stringWithUTF8String:command], @"command", (args ? (id) [NSString stringWithUTF8String:args] : (id) [NSNull null]), @"arguments", nil]];
}

#pragma mark -

@implementation MVChatConnection
- (id) init {
	self = [super init];

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection init]\n");
#endif

    _status = MVChatConnectionDisconnectedStatus;
	_chatConnection = firetalk_create_handle( FP_IRC, (void *) self );
	_joinList = [[NSMutableArray array] retain];
	_roomsCache = [[NSMutableDictionary dictionary] retain];

	[self setFloodControlIntervals:[NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithDouble:6.], @"messages", [NSNumber numberWithDouble:1.], @"delay", [NSNumber numberWithDouble:1.5], @"factor", [NSNumber numberWithDouble:3.], @"ceiling", nil]];

	[self _registerCallbacks];
	[self _registerForSleepNotifications];

	_firetalkSelectTimer = [[NSTimer scheduledTimerWithTimeInterval:.100 target:self selector:@selector( _executeRunLoopCheck: ) userInfo:nil repeats:YES] retain];

	return self;
}

- (id) initWithURL:(NSURL *) url {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection initWithURL]\n");
#endif

    if( ! [[url scheme] isEqualToString:@"irc"] ) return nil;
    
    self = [self initWithServer:[url host] port:[[url port] unsignedShortValue] user:[url user] password:[url password]];

    return self;
}

- (id) initWithServer:(NSString *) server port:(unsigned short) port user:(NSString *) nickname password:(NSString *)password {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection initWithServer]\n");
#endif

    self = [self init];

    [self setPassword:password];
	[self setNickname:nickname];
	[self setServer:server];
	[self setServerPort:port];

    return self;
}

#if 0 //Moved all to -dealloc
- (void) release {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection release]\n");
#endif

    if( ( [self retainCount] - 1 ) == 1 ) {
		[_firetalkSelectTimer invalidate];
	}

    [super release];
}
#endif

- (void) dealloc {

    [self _deregisterForSleepNotifications];
    [_firetalkSelectTimer invalidate]; //Markus: moved here from -release
	[_firetalkSelectTimer autorelease];

	[_nickname autorelease];
	[_password autorelease];
	[_server autorelease];
	[_joinList autorelease];
	[_roomsCache autorelease];

	[[NSNotificationCenter defaultCenter] removeObserver:self];

	firetalk_destroy_handle( _chatConnection );
	_chatConnection = NULL;
	_nickname = nil;
	_password = nil;
	_server = nil;
	_joinList = nil;
	_roomsCache = nil;
	_firetalkSelectTimer = nil;

	[super dealloc];
}

#pragma mark -

- (void) connect {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection connect]\n");
    printf("Server %s  Port %d  Nick %s  Passwort %s\n", [_server UTF8String], _port, [_nickname UTF8String], [_password UTF8String]);
#endif

    if( ! _server ) return;
	if( ! _nickname ) return;
	if( [self status] == MVChatConnectionConnectingStatus ) return;
	if( [self isConnected] ) [self disconnect];
	[self _willConnect];

    firetalk_set_password( _chatConnection, [_password UTF8String]);
	firetalk_signon( _chatConnection, [_server UTF8String], _port, [_nickname UTF8String] );
}

- (void) connectToServer:(NSString *) server onPort:(unsigned short) port asUser:(NSString *) nickname withPassword:(NSString *) password {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection connectToServer]\n");
#endif

    if( [self isConnected] ) [self disconnect];

    [self setPassword:password];
    [self setNickname:nickname];
	[self setServer:server];
	[self setServerPort:port];

    [self connect];
}

- (void) disconnect {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection disconnect]\n");
#endif

    if( [self status] != MVChatConnectionDisconnectedStatus ) {
		[self _willDisconnect];
		firetalk_disconnect( _chatConnection );
	}
}

#pragma mark -

- (NSURL *) url {
	NSString *url = nil;

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection url]\n");
#endif

    if( ! _server ) return nil;
	if( _nickname && _port ) url = [NSString stringWithFormat:@"irc://%@@%@:%hu", MVURLEncodeString( _nickname ), MVURLEncodeString( _server ), _port];
	else if( _nickname && ! _port ) url = [NSString stringWithFormat:@"irc://%@@%@", MVURLEncodeString( _nickname ), MVURLEncodeString( _server )];
	else url = [NSString stringWithFormat:@"irc://%@", MVURLEncodeString( _server )];
	return [[[NSURL URLWithString:url] retain] autorelease];
}

#pragma mark -

- (void) setNickname:(NSString *) nickname {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection setNickname]\n");
#endif
    NSParameterAssert([nickname length]);

    if( [self isConnected] && ! [nickname isEqualToString:_nickname] ) {
        firetalk_set_nickname( _chatConnection, [nickname UTF8String] );
    }
    [_nickname autorelease];
    _nickname = [nickname copy];
}

- (NSString *) nickname {
	return [[_nickname copy] autorelease];
}

#pragma mark -

/* Markus
- (void) setNicknamePassword:(NSString *) password {
	if( [self isConnected] && [password length] && ! [password isEqualToString:_npassword] ) {
		firetalk_set_password( _chatConnection, [password UTF8String] );
	}
	[_npassword autorelease];
	if( [password length] ) _npassword = [password copy];
	else _npassword = nil;
}

- (NSString *) nicknamePassword {
	return [[_npassword copy] autorelease];
}
*/

#pragma mark -

- (void) setPassword:(NSString *) password {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection setPassword]\n");
#endif

    if( ! [self isConnected] && [password length] && ! [password isEqualToString:_password] ) {
		firetalk_set_password( _chatConnection, [password UTF8String] );
	}
	[_password autorelease];
	if( [password length] ) _password = [password copy];
	else _password = nil;
}

- (NSString *) password {
	return [[_password copy] autorelease];
}

#pragma mark -

- (void) setServer:(NSString *) server {
    [_server autorelease];
	_server = [server copy];
}

- (NSString *) server {
	return [[_server copy] autorelease];
}

#pragma mark -

- (void) setServerPort:(unsigned short) port {
	_port = ( port ? port : 6667 );
}

- (unsigned short) serverPort {
	return _port;
}

#pragma mark -

- (NSDictionary *) floodControlIntervals {
	return [[_floodIntervals retain] autorelease];
}

- (void) setFloodControlIntervals:(NSDictionary *) intervals {
	firetalk_set_flood_intervals( _chatConnection, [[intervals objectForKey:@"messages"] doubleValue], [[intervals objectForKey:@"delay"] doubleValue], [[intervals objectForKey:@"factor"] doubleValue], [[intervals objectForKey:@"ceiling"] doubleValue] );
	[_floodIntervals autorelease];
	_floodIntervals = [intervals copy];
}

#pragma mark -

- (void) sendMessageToUser:(NSString *) user attributedMessage:(NSAttributedString *) message withEncoding:(NSStringEncoding) encoding asAction:(BOOL) action {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection sendMessageToUser]\n");
#endif

    if( [self isConnected] ) {
		MVChatWindowController *window = [MVChatWindowController chatWindowWithUser:user withConnection:self ifExists:YES];
		NSMutableData *encodedData = [[[MVChatWindowController flattenedHTMLDataForMessage:message withEncoding:encoding] mutableCopy] autorelease];

		[window addHTMLMessageToDisplay:encodedData fromUser:[self nickname] asAction:action asAlert:NO];

		[encodedData appendBytes:"\0" length:1];
		if( action ) firetalk_im_send_action( _chatConnection, [user UTF8String], (char *) [encodedData bytes], 0 );
		else firetalk_im_send_message( _chatConnection, [user UTF8String], (char *) [encodedData bytes], 0 );
		
		[MVChatWindowController updateChatWindowsMember:_nickname withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
	} else [[MVChatWindowController chatWindowWithUser:user withConnection:self ifExists:YES] disconnected];
}

- (void) sendMessageToChatRoom:(NSString *) room attributedMessage:(NSAttributedString *) message withEncoding:(NSStringEncoding) encoding asAction:(BOOL) action {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection sendMessageToChatRoom]\n");
#endif

    if( [self isConnected] ) {
		MVChatWindowController *window = [MVChatWindowController chatWindowForRoom:room withConnection:self ifExists:YES];
		NSMutableData *encodedData = [[[MVChatWindowController flattenedHTMLDataForMessage:message withEncoding:encoding] mutableCopy] autorelease];

		[window addHTMLMessageToDisplay:encodedData fromUser:[self nickname] asAction:action asAlert:NO];

		[encodedData appendBytes:"\0" length:1];
		if( action ) firetalk_chat_send_action( _chatConnection, [[room lowercaseString] UTF8String], (char *) [encodedData bytes], 0 );
		else firetalk_chat_send_message( _chatConnection, [[room lowercaseString] UTF8String], (char *) [encodedData bytes], 0 );

		[MVChatWindowController updateChatWindowsMember:_nickname withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
	} else [[MVChatWindowController chatWindowForRoom:[room lowercaseString] withConnection:self ifExists:YES] disconnected];
}

- (void) sendRawMessage:(NSString *) text {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection sendRawMessage]\n");
#endif
    
    firetalk_send_raw( _chatConnection, [text UTF8String]);
}

#pragma mark -

- (void) sendFileToUser:(NSString *) user withFilePath:(NSString *) path {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection sendFileToUser]\n");
#endif

    if( ! [[NSFileManager defaultManager] isReadableFileAtPath:path] ) return;
	if( [self isConnected] ) {
		NSNumber *size = [[[NSFileManager defaultManager] fileAttributesAtPath:path traverseLink:YES] objectForKey:@"NSFileSize"];
		void *handle = NULL;
		firetalk_file_offer( _chatConnection, &handle, [user UTF8String], [path fileSystemRepresentation] );
		[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionFileTransferOfferedNotification object:self userInfo:[NSDictionary dictionaryWithObjectsAndKeys:[NSString stringWithFormat:@"%x", handle], @"identifier", user, @"to", path, @"path", size, @"size", nil]];
	}
}

- (void) acceptFileTransfer:(NSString *) identifier saveToPath:(NSString *) path resume:(BOOL) resume  {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection acceptFileTransfer]\n");
#endif

    if( [self isConnected] ) {
		void *pointer = NULL;
		sscanf( [identifier UTF8String], "%8lx", (unsigned long int *) &pointer );
		if( resume ) firetalk_file_resume( _chatConnection, pointer, NULL, [path fileSystemRepresentation] );
		else firetalk_file_accept( _chatConnection, pointer, NULL, [path fileSystemRepresentation] );
	}
}

- (void) cancelFileTransfer:(NSString *) identifier {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection cancelFileTransfer]\n");
#endif

    if( [self isConnected] ) {
		void *pointer = NULL;
		sscanf( [identifier UTF8String], "%8lx", (unsigned long int *) &pointer );
		firetalk_file_cancel( _chatConnection, pointer );
	}
}

#pragma mark -

- (void) sendSubcodeRequest:(NSString *) command toUser:(NSString *) user withArguments:(NSString *) arguments {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection sendSubcodeRequest]\n");
#endif

    NSParameterAssert( command != nil );
	NSParameterAssert( user != nil );
	if( [self isConnected] ) {
		firetalk_subcode_send_request( _chatConnection, [user UTF8String], [command UTF8String], [arguments UTF8String] );
	}
}

- (void) sendSubcodeReply:(NSString *) command toUser:(NSString *) user withArguments:(NSString *) arguments {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection sendSubcodeReply]\n");
#endif

    NSParameterAssert( command != nil );
	NSParameterAssert( user != nil );
	if( [self isConnected] ) {
		firetalk_subcode_send_reply( _chatConnection, [user UTF8String], [command UTF8String], [arguments UTF8String] );
	}
}

#pragma mark -

- (void) joinChatForRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection joinChatForRoom]\n");
#endif

    if( [self isConnected] ) firetalk_chat_join( _chatConnection, [[room lowercaseString] UTF8String] );
	else [_joinList addObject:room];
}

- (void) partChatForRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection partChatForRoom]\n");
#endif

    if( [self isConnected] ) firetalk_chat_part( _chatConnection, [[room lowercaseString] UTF8String] );
}

#pragma mark -

- (void) setTopic:(NSAttributedString *) topic withEncoding:(NSStringEncoding) encoding forRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection setTopic]\n");
#endif

    NSParameterAssert( room != nil );
	if( [self isConnected] ) {
		NSData *encodedData = [MVChatWindowController flattenedHTMLDataForMessage:topic withEncoding:encoding];
		firetalk_chat_set_topic( _chatConnection, [[room lowercaseString] UTF8String], (char *) [encodedData bytes] );
		[MVChatWindowController updateChatWindowsMember:_nickname withInfo:[NSDictionary dictionaryWithObject:[NSNumber numberWithUnsignedInt:0] forKey:@"idle"] forConnection:self];
	}
}

#pragma mark -

- (void) promoteMember:(NSString *) member inRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection promoteMember]\n");
#endif

    NSParameterAssert( member != nil );
	NSParameterAssert( room != nil );
	if( [self isConnected] ) firetalk_chat_op( _chatConnection, [[room lowercaseString] UTF8String], [member UTF8String] );
}

- (void) demoteMember:(NSString *) member inRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection demoteMember]\n");
#endif

    NSParameterAssert( member != nil );
	NSParameterAssert( room != nil );
	if( [self isConnected] ) firetalk_chat_deop( _chatConnection, [[room lowercaseString] UTF8String], [member UTF8String] );
}

- (void) voiceMember:(NSString *) member inRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection voiceMember]\n");
#endif

    NSParameterAssert( member != nil );
	NSParameterAssert( room != nil );
	if( [self isConnected] ) firetalk_chat_voice( _chatConnection, [[room lowercaseString] UTF8String], [member UTF8String] );
}

- (void) devoiceMember:(NSString *) member inRoom:(NSString *) room {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection devoiceMember]\n");
#endif

    NSParameterAssert( member != nil );
	NSParameterAssert( room != nil );
	if( [self isConnected] ) firetalk_chat_devoice( _chatConnection, [[room lowercaseString] UTF8String], [member UTF8String] );
}

- (void) kickMember:(NSString *) member inRoom:(NSString *) room forReason:(NSString *) reason {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection kickMember]\n");
#endif

    NSParameterAssert( member != nil );
	NSParameterAssert( room != nil );
	if( [self isConnected] ) firetalk_chat_kick( _chatConnection, [[room lowercaseString] UTF8String], [member UTF8String], [reason UTF8String] );
}

#pragma mark -

- (void) addUserToNotificationList:(NSString *) user {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection addUserToNotificationList]\n");
#endif

    NSParameterAssert( user != nil );
	firetalk_im_internal_add_buddy( _chatConnection, [user UTF8String] );
}

- (void) removeUserFromNotificationList:(NSString *) user {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection removeUserFromNotificationList]\n");
#endif

    NSParameterAssert( user != nil );
	firetalk_im_internal_remove_buddy( _chatConnection, [user UTF8String] );
}

#pragma mark -

- (void) fetchInformationForUser:(NSString *) user withPriority:(BOOL) priority {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection fetchInformationForUser]\n");
#endif

    NSParameterAssert( user != nil );
	if( [self isConnected] ) firetalk_im_get_info( _chatConnection, [user UTF8String], priority );
}

#pragma mark -

- (void) fetchRoomList {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection fetchRoomList]\n");
#endif

    if( [self isConnected] ) {
		if( ! _cachedDate || [_cachedDate timeIntervalSinceNow] < -900. ) {
			firetalk_im_get_roomlist( _chatConnection, NULL );
			[_cachedDate autorelease];
			_cachedDate = [[NSDate date] retain];
		}
	}
}

- (void) fetchRoomListWithRooms:(NSArray *) rooms {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection fetchRoomListWithRooms]\n");
#endif

    if( [self isConnected] ) {
		NSString *search = [rooms componentsJoinedByString:@","];
		firetalk_im_get_roomlist( _chatConnection, [search UTF8String] );
	}
}

- (void) stopFetchingRoomList {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection stopFetchingRoomList]\n");
#endif

    if( [self isConnected] ) {
		firetalk_im_stop_roomlist( _chatConnection );
	}
}

- (NSDictionary *) roomListResults {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection roomListResults]\n");
#endif

    return [[_roomsCache retain] autorelease];
}

#pragma mark -

- (void) setAwayStatusWithMessage:(NSString *) message {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection setAwayStatusWithMessage]\n");
#endif

    if( [self isConnected] ) {
		if( [message length] ) firetalk_set_away( _chatConnection, [message UTF8String] );
		else firetalk_set_away( _chatConnection, NULL );
	}
}

- (void) clearAwayStatus {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection clearAwayStatus]\n");
#endif

    [self setAwayStatusWithMessage:nil];
}

#pragma mark -

- (BOOL) isConnected {
	return (BOOL) ( _status == MVChatConnectionConnectedStatus );
}

- (MVChatConnectionStatus) status {
	return _status;
}
@end

#pragma mark -

@implementation MVChatConnection (MVChatConnectionPrivate)
- (io_connect_t) _powerConnection {
	return _powerConnection;
}

- (firetalk_t) _firetalkConnection {
	return _chatConnection;
}

#pragma mark -

- (void) _executeRunLoopCheck:(NSTimer *) timer {
	struct timeval timeout = { 0, 050 };
//Markus: is this one really necessary?
    
    if( firetalk_select_custom( 0, NULL, NULL, NULL, &timeout ) < 0 ) {
		NSLog( @"firetalk_select: %s", firetalk_strerror( firetalkerror ) );
	}
}

- (void) _registerCallbacks {
	if( ! _chatConnection ) return;
	firetalk_register_callback( _chatConnection, FC_DOINIT, (firetalk_callback) MVChatLoad );
	firetalk_register_callback( _chatConnection, FC_CONNECTED, (firetalk_callback) MVChatConnected );
	firetalk_register_callback( _chatConnection, FC_CONNECTFAILED, (firetalk_callback) MVChatConnectionFailed );
	firetalk_register_callback( _chatConnection, FC_DISCONNECT, (firetalk_callback) MVChatDisconnect );
	firetalk_register_callback( _chatConnection, FC_ERROR, (firetalk_callback) MVChatError );
	firetalk_register_callback( _chatConnection, FC_BACKLOG, (firetalk_callback) MVChatBackLog );
	firetalk_register_callback( _chatConnection, FC_NEWNICK, (firetalk_callback) MVChatNewNickname );
	firetalk_register_callback( _chatConnection, FC_NEEDPASS, (firetalk_callback) MVChatNeedPassword );
	firetalk_register_callback( _chatConnection, FC_IM_GETMESSAGE, (firetalk_callback) MVChatGetMessage );
	firetalk_register_callback( _chatConnection, FC_IM_GETACTION, (firetalk_callback) MVChatGetAction );
	firetalk_register_callback( _chatConnection, FC_IM_BUDDYONLINE, (firetalk_callback) MVChatBuddyOnline );
	firetalk_register_callback( _chatConnection, FC_IM_BUDDYOFFLINE, (firetalk_callback) MVChatBuddyOffline );
	firetalk_register_callback( _chatConnection, FC_IM_BUDDYAWAY, (firetalk_callback) MVChatBuddyAway );
	firetalk_register_callback( _chatConnection, FC_IM_BUDDYUNAWAY, (firetalk_callback) MVChatBuddyUnaway );
	firetalk_register_callback( _chatConnection, FC_IM_LISTBUDDY, (firetalk_callback) MVChatListBuddy );
	firetalk_register_callback( _chatConnection, FC_IM_LISTROOM, (firetalk_callback) MVChatListRoom );
	firetalk_register_callback( _chatConnection, FC_IM_GOTINFO, (firetalk_callback) MVChatGotInfo );
	firetalk_register_callback( _chatConnection, FC_IM_IDLEINFO, (firetalk_callback) MVChatBuddyGotIdle );
	firetalk_register_callback( _chatConnection, FC_CHAT_JOINED, (firetalk_callback) MVChatJoinedRoom );
	firetalk_register_callback( _chatConnection, FC_CHAT_LEFT, (firetalk_callback) MVChatLeftRoom );
	firetalk_register_callback( _chatConnection, FC_CHAT_KICKED, (firetalk_callback) MVChatKicked );
	firetalk_register_callback( _chatConnection, FC_CHAT_GETMESSAGE, (firetalk_callback) MVChatGetRoomMessage );
	firetalk_register_callback( _chatConnection, FC_CHAT_GETACTION, (firetalk_callback) MVChatGetRoomAction );
	firetalk_register_callback( _chatConnection, FC_CHAT_INVITED, (firetalk_callback) MVChatInvited );
	firetalk_register_callback( _chatConnection, FC_CHAT_GOTTOPIC, (firetalk_callback) MVChatGotTopic );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_JOINED, (firetalk_callback) MVChatUserJoinedRoom );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_LEFT, (firetalk_callback) MVChatUserLeftRoom );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_OPPED, (firetalk_callback) MVChatUserOpped );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_DEOPPED, (firetalk_callback) MVChatUserDeopped );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_VOICED, (firetalk_callback) MVChatUserVoiced );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_DEVOICED, (firetalk_callback) MVChatUserDevoiced );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_KICKED, (firetalk_callback) MVChatUserKicked );
	firetalk_register_callback( _chatConnection, FC_CHAT_USER_NICKCHANGED, (firetalk_callback) MVChatUserNicknameChanged );
	firetalk_register_callback( _chatConnection, FC_FILE_OFFER, (firetalk_callback) MVChatFileTransferAccept );
	firetalk_register_callback( _chatConnection, FC_FILE_START, (firetalk_callback) MVChatFileTransferStart );
	firetalk_register_callback( _chatConnection, FC_FILE_FINISH, (firetalk_callback) MVChatFileTransferFinish );
	firetalk_register_callback( _chatConnection, FC_FILE_ERROR, (firetalk_callback) MVChatFileTransferError );
	firetalk_register_callback( _chatConnection, FC_FILE_PROGRESS, (firetalk_callback) MVChatFileTransferStatus );
	firetalk_subcode_register_request_callback( _chatConnection, "VERSION", (firetalk_subcode_callback) MVChatSubcodeRequest );
	firetalk_subcode_register_request_callback( _chatConnection, "USERINFO", (firetalk_subcode_callback) MVChatSubcodeRequest );
	firetalk_subcode_register_request_callback( _chatConnection, "URL", (firetalk_subcode_callback) MVChatSubcodeRequest );
	firetalk_subcode_register_request_callback( _chatConnection, NULL, (firetalk_subcode_callback) MVChatSubcodeRequest );
	firetalk_subcode_register_reply_callback( _chatConnection, NULL, (firetalk_subcode_callback) MVChatSubcodeReply );
}

#pragma mark -

- (void) _registerForSleepNotifications {
	IONotificationPortRef sleepNotePort = NULL;
	CFRunLoopSourceRef rls = NULL;
	_powerConnection = IORegisterForSystemPower( (void *) self, &sleepNotePort, MVChatHandlePowerChange, &_sleepNotifier );
	if( ! _powerConnection ) return;
	rls = IONotificationPortGetRunLoopSource( sleepNotePort );
	CFRunLoopAddSource( CFRunLoopGetCurrent(), rls, kCFRunLoopDefaultMode );
	CFRelease( rls );
}

- (void) _deregisterForSleepNotifications {
	IODeregisterForSystemPower( &_sleepNotifier );
	_powerConnection = NULL;
}

#pragma mark -

- (void) _addRoomToCache:(NSString *) room withUsers:(int) users andTopic:(NSString *) topic {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _addRoomToCache]\n");
#endif

    if( room ) {
		NSDictionary *info = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithUnsignedInt:users], @"users", ( topic ? (id) topic : (id) [NSNull null] ), @"topic", [NSDate date], @"cached", nil];
		[_roomsCache setObject:info forKey:room];
	}
}

- (void) _setBacklogDelay:(NSTimeInterval) delay {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _setBacklogDelay]\n");
#endif

    _backlogDelay = delay;
}

#pragma mark -

- (void) _setStatus:(MVChatConnectionStatus) status {
	_status = status;
}

- (void) _willConnect {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _willConnect]\n");
#endif

    _status = MVChatConnectionConnectingStatus;
// unused	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionWillConnectNotification object:self];
}

- (void) _didConnect {
	NSEnumerator *enumerator = [_joinList objectEnumerator];
	NSString *room = nil;


#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _didConnect]\n");
#endif

    _status = MVChatConnectionConnectedStatus;
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionDidConnectNotification object:self];

	while( ( room = [enumerator nextObject] ) )
		[self joinChatForRoom:room];

	[_joinList removeAllObjects];
}

- (void) _didNotConnect {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _didNotConnect]\n");
#endif

    _status = MVChatConnectionDisconnectedStatus;
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionDidNotConnectNotification object:self];
}

- (void) _willDisconnect {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _willDisconnect]\n");
#endif

    [[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionWillDisconnectNotification object:self];
}

- (void) _didDisconnect {

#ifdef DEBUG_CHAT_CONNECTION
    printf("-[MVChatConnection _didDisconnect]\n");
#endif

    _status = MVChatConnectionDisconnectedStatus;
	[[NSNotificationCenter defaultCenter] postNotificationName:MVChatConnectionDidDisconnectNotification object:self];
}
@end
