/*
irc.c - FireTalk IRC protocol definitions
Copyright (C) 2000 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>
#include <time.h>
#include <ctype.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "firetalk.h"
#include "firetalk-int.h"
#include "irc.h"

#define IRC_SERVER "irc.openprojects.net"
#define IRC_PORT 6667

typedef struct s_irc_whois {
	struct s_irc_whois *next;
	char *nickname;
	char *info;
	int flags;
	long idle;
} s_irc_whois;

typedef struct s_irc_connection {
	int s;
	char *nickname;
	char *password;
	char buffer[513];
	time_t lasttime;                                     /* last time we set idles */
	int isons;                                           /* number of ISON's we're waiting on replies to */
	struct s_irc_whois *whois_head;
	int passchange;                                      /* whether we are currently changing our pass */
	int usesilence;                                      /* are we on a network that understands SILENCE */
	int identified;                                      /* are we identified */
	int color_filter;                                    /* to filter out irc colors in irc_irc_to_html() */
	struct s_irc_host_cache *host_cache;				 /* a chae tree to store all users that cme in some contact with you (used by DCC) */
	struct s_irc_dcc_connection *dcc_active;
} s_irc_connection;

typedef struct s_irc_dcc_connection {
	struct s_irc_connection *irc;
	char *filename;
	char *user;
	char *path;
	u_long size;
	u_long position;
	u_long host;
	u_short port;
	pthread_t thread;
	char finished;
	int ifd, ofd, socket;
	struct s_irc_dcc_connection *back, *next;
} s_irc_dcc_connection;

typedef struct s_irc_host_cache {
	char *nickname;
	char *address;
	u_long host;
	struct s_irc_host_cache *back, *next;
} s_irc_host_cache;

static char *irc_html_to_irc(const char * const string);
static char *irc_irc_to_html(const char * const string, const int filter_colors);
static int irc_send_printf(struct s_irc_connection * const c, const char * const format, ...);
static int irc_internal_disconnect(struct s_irc_connection * const c, const int error);
static char **irc_recv_parse(struct s_irc_connection * const c);

void *irc_dcc_send_file(void *c);
void *irc_dcc_receive_file(void *c);

void irc_dcc_conenction_add(struct s_irc_connection * const conn,struct s_irc_dcc_connection * const dcc);
void irc_dcc_conenction_remove(struct s_irc_connection * const conn,struct s_irc_dcc_connection * const dcc);
s_irc_dcc_connection * const irc_dcc_conenction_find_by_port(struct s_irc_connection * const conn,const u_short port);
s_irc_dcc_connection * const irc_dcc_conenction_find_by_names(struct s_irc_connection * const conn,const char * const who,const char * const file);
void irc_dcc_conenction_cleanup(struct s_irc_connection * const conn);

void irc_host_cache_add(struct s_irc_connection * const conn,const char * const hostmask);
void irc_host_cache_remove(struct s_irc_connection * const conn,const char * const hostmask);
void irc_host_cache_change(struct s_irc_connection * const conn,const char * const old,const char * const new);
struct s_irc_host_cache * const irc_host_cache_find(struct s_irc_connection * const conn,const char * const user);
void irc_host_cache_free(struct s_irc_connection * const conn);

char *name_with_underscores( const char *name );
char *name_with_spaces( const char *name );

/*

from buddy structure:

tempint1 = ison, default 0, used during ISON checks
tempint2 = away, default 0, used during WHOIS checks

*/

static const char *irc_irc_to_html_colors[] = { // matches mIRC's 16 colors
	"#FFFFFF",
	"#000000",
	"#000066",
	"#009900",
	"#FF0000",
	"#660000",
	"#990099",
	"#FF6600",
	"#FFFF00",
	"#00FF00",
	"#009999",
	"#00FFFF",
	"#0000FF",
	"#FF00FF",
	"#666666",
	"#CCCCCC"
};

static char *irc_html_to_irc(const char * const string) {
	static char *output = NULL;
	int l = 0,i = 0,o = 0, incolor = 0;
	unsigned color = 0;
	if( string ) l = strlen(string);
	else l = 0;
	output = realloc(output,l + 1);
	memset( output, 0, l + 1 );
	while (i < l) {
		switch(string[i]) {
			case '&':
				if (!strncasecmp(&string[i],"&amp;",5)) {
					output[o++] = '&';
					i += 5;
				} else if (!strncasecmp(&string[i],"&gt;",4)) {
					output[o++] = '>';
					i += 4;
				} else if (!strncasecmp(&string[i],"&lt;",4)) {
					output[o++] = '<';
					i += 4;
				} else
					output[o++] = string[i++];
				break;
			case '<':
				if (!strncasecmp(&string[i],"<b>",3)) {
					output[o++] = 2;
					i += 3;
				} else if (!strncasecmp(&string[i],"</b>",4)) {
					output[o++] = 2;
					i += 4;
				} else if (!strncasecmp(&string[i],"<i>",3)) {
					output[o++] = 22;
					i += 3;
				} else if (!strncasecmp(&string[i],"</i>",4)) {
					output[o++] = 22;
					i += 4;
				} else if (!strncasecmp(&string[i],"<u>",3)) {
					output[o++] = 31;
					i += 3;
				} else if (!strncasecmp(&string[i],"</u>",4)) {
					output[o++] = 31;
					i += 4;
				} else if (!strncasecmp(&string[i],"<a href=",8)) {
					if( string[i + 8] == '"' || string[i + 8] == '\'' ) i += 9;
					else i += 8;
					output[o++] = 31;
					while( i < l && string[i] != '"' && string[i] != '\'' )
						output[o++] = string[i++];
					while( i < l && strncasecmp(&string[i],"</a>",4) ) i++;
					output[o++] = 31;
					i += 4;
				} else if (!strncasecmp(&string[i],"<font",5)) {
					int r = 0, g = 0, b = 0;
					char colorCode[2];
					while( i < l && strncasecmp(&string[i],"color=",6) ) i++;
					if( string[i + 6] == '"' || string[i + 6] == '\'' ) i += 7;
					else i += 6;
					if( string[i] == '#' ) i += 1;
					sscanf( &string[i], "%x", &color );
					while( i < l && strncasecmp(&string[i],">",1) ) i++;
					i += 1;
					r = (color & 0xff0000) >> 16;
					g = (color & 0x00ff00) >> 8;
					b = (color & 0x0000ff) >> 0;
					// Kind of messy but this picks the closest mIRC color match. (24bit color -> 4bit mIRC color)
					// NOTE: not all colors will match (only relitivly close ones), no color is sent if a close match can't be found.
					if( r == 0xff && g == 0xff && b == 0xff ) { colorCode[0] = '0'; colorCode[1] = '0'; }
					else if( r == 0x00 && g == 0x00 && b == 0x00 ) { colorCode[0] = '0'; colorCode[1] = '1'; }
					else if( r < 0x40 && g < 0x40 && b < 0xB0 && b > 0x40 ) { colorCode[0] = '0'; colorCode[1] = '2'; }
					else if( r < 0x40 && g < 0xC0 && g > 0x70 && b < 0x40 ) { colorCode[0] = '0'; colorCode[1] = '3'; }
					else if( r <= 0xff && r > 0xC0 && g < 0x40 && b < 0x40 ) { colorCode[0] = '0'; colorCode[1] = '4'; }
					else if( r < 0xC0 && r > 0x40 && g < 0x40 && b < 0x40 ) { colorCode[0] = '0'; colorCode[1] = '5'; }
					else if( r < 0xC0 && r > 0x70 && g < 0x40 && b < 0xC0 && b > 0x70 ) { colorCode[0] = '0'; colorCode[1] = '6'; }
					else if( r <= 0xff && r > 0xC0 && g < 0xB0 && g > 0x40 && b < 0x40 ) { colorCode[0] = '0'; colorCode[1] = '7'; }
					else if( r <= 0xff && r > 0xC0 && g <= 0xff && g > 0xC0 && b < 0x40 ) { colorCode[0] = '0'; colorCode[1] = '8'; }
					else if( r < 0x40 && g <= 0xff && g > 0xC0 && b < 0x40 ) { colorCode[0] = '0'; colorCode[1] = '9'; }
					else if( r < 0x40 && g < 0xC0 && g > 0x70 && b < 0xC0 && b > 0x70 ) { colorCode[0] = '1'; colorCode[1] = '0'; }
					else if( r < 0x40 && g <= 0xff && g > 0xC0 && b <= 0xff && b > 0xC0 ) { colorCode[0] = '1'; colorCode[1] = '1'; }
					else if( r < 0x40 && g < 0x40 && b <= 0xff && b > 0xC0 ) { colorCode[0] = '1'; colorCode[1] = '2'; }
					else if( r <= 0xff && r > 0xC0 && g < 0x40 && b <= 0xff && b > 0xC0 ) { colorCode[0] = '1'; colorCode[1] = '3'; }
					else if( r < 0x80 && r > 0x40 && g < 0x80 && g > 0x40 && b < 0x80 && b > 0x40 ) { colorCode[0] = '1'; colorCode[1] = '4'; }
					else if( r < 0xe0 && r > 0x80 && g < 0xe0 && g > 0x80 && b < 0xe0 && b > 0x80 ) { colorCode[0] = '1'; colorCode[1] = '5'; }
					else { colorCode[0] = 0; colorCode[1] = 0; }
					if( colorCode[0] ) {
						incolor = 1;
						output[o++] = 3;
						output[o++] = colorCode[0];
						output[o++] = colorCode[1];
					} else incolor = 0;
				} else if (!strncasecmp(&string[i],"</font>",7)) {
					if( incolor ) output[o++] = 3;
					i += 7;
				} else
					output[o++] = string[i++];
				break;
			default:
				output[o++] = string[i++];
				break;
		}
	}
	output[o] = '\0';
	return output;
}

static char *irc_irc_to_html(const char * const string, const int filter_colors) {
	static char *output = NULL;
	int l,i = 0,o = 0;
	int inbold = 0, initalics = 0, inunderline = 0, incolor = 0;
	int colorCode = -1, backColorCode = -1;
	l = strlen(string);
	output = realloc(output,(l * 5) + 1);
	while (i < l) {
		switch(string[i]) {
			case 2:
				if (inbold) {
					memcpy(&output[o],"</b>",4);
					o += 4;
					inbold = 0;
				} else {
					memcpy(&output[o],"<b>",3);
					o += 3;
					inbold = 1;
				}
				break;
			case 3:
				if( isdigit( string[i + 1] ) ) {
					if( isdigit( string[i + 2] ) ) {
						colorCode = ( string[i + 1] - '0' ) * 10;
						colorCode += ( string[i + 2] - '0' );
						if( string[i + 3] == ',' ) {
							if( isdigit( string[i + 4] ) ) {
								if( isdigit( string[i + 5] ) ) {
									backColorCode = ( string[i + 4] - '0' ) * 10;
									backColorCode += ( string[i + 5] - '0' );
									i += 1;
								} else backColorCode = ( string[i + 4] - '0' );
								i += 1;
							}
							i += 1;
						}
						i += 1;
					} else if( string[i + 2] == ',' ) {
						colorCode = ( string[i + 1] - '0' );
						if( isdigit( string[i + 3] ) ) {
							if( isdigit( string[i + 4] ) ) {
								backColorCode = ( string[i + 3] - '0' ) * 10;
								backColorCode += ( string[i + 4] - '0' );
								i += 1;
							} else backColorCode = ( string[i + 3] - '0' );
							i += 1;
						}
						i += 1;
					} else colorCode = ( string[i + 1] - '0' );
					if( backColorCode > 15 ) backColorCode = -1;
					i += 1;
					/*if( incolor ) {
						memcpy(&output[o],"</font>",7);
						o += 7;
					}*/
					if( colorCode <= 15 && colorCode >= 0 && ! filter_colors ) {
						memcpy(&output[o],"<font color=\"",13);
						o += 13;
						memcpy(&output[o],irc_irc_to_html_colors[colorCode],7);
						o += 7;
						if( backColorCode <= 15 && backColorCode >= 0 ) {
							memcpy(&output[o],"\" style=\"background-color:",26);
							o += 26;
							memcpy(&output[o],irc_irc_to_html_colors[backColorCode],7);
							o += 7;
						}
						memcpy(&output[o],"\">",2);
						o += 2;
						incolor = 1;
					}
					colorCode = -1;
					//backColorCode = -1;
				} else if( incolor && ! filter_colors ) {
					memcpy(&output[o],"</font>",7);
					o += 7;
				}
				break;
			case 22:
				if (initalics) {
					memcpy(&output[o],"</i>",4);
					o += 4;
					initalics = 0;
				} else {
					memcpy(&output[o],"<i>",3);
					o += 3;
					initalics = 1;
				}
				break;
			case 31:
				if (inunderline) {
					memcpy(&output[o],"</u>",4);
					o += 4;
					inunderline = 0;
				} else {
					memcpy(&output[o],"<u>",3);
					o += 3;
					inunderline = 1;
				}
				break;
			case '&':
				memcpy(&output[o],"&amp;",5);
				o += 5;
				break;
			case '<':
				memcpy(&output[o],"&lt;",4);
				o += 4;
				break;
			case '>':
				memcpy(&output[o],"&gt;",4);
				o += 4;
				break;
			default:
				output[o++] = string[i];
				break;
		}
		i++;
	}
	output[o] = '\0';
	return output;
}

static int irc_internal_disconnect(struct s_irc_connection * const c, const int error) {
	shutdown(c->s, SHUT_RDWR);
	if (c->nickname)
		free(c->nickname);
	if (c->password)
		free(c->password);
	irc_host_cache_free(c);
	firetalk_callback_disconnect(c,error);
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

static int irc_send_printf(struct s_irc_connection * const c, const char * const format, ...) {
	va_list ap;
	int len;
	int i;
	char data[513];
	short datai = 0;
	char *tempchr;

	va_start(ap,format);

	len = strlen(format);
	for (i = 0; i < len; i++) {
		if (format[i] == '%') {
			switch (format[++i]) {
				case 's':
					tempchr = irc_html_to_irc(va_arg(ap, char *));
					if (datai + strlen(tempchr) > 509) {
						firetalkerror = FE_PACKETSIZE;
						return FE_PACKETSIZE;
					}
					strcpy(&data[datai],tempchr);
					datai += strlen(tempchr);
					break;
				case '%':
					data[datai++] = '%';
					break;
			}
		} else {
			data[datai++] = format[i];
			if (datai > 509) {
				firetalkerror = FE_PACKETSIZE;
				return FE_PACKETSIZE;
			}
		}
	}
	data[datai] = '\0';

	strcat(data,"\r\n");
	datai = strlen(data);

	if (send(c->s,data,datai,0) != datai) {
		irc_internal_disconnect(c,FE_PACKET);
		firetalkerror = FE_PACKET;
		return FE_PACKET;
	}
	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

static char **irc_recv_parse(struct s_irc_connection * const c) {
	static char *args[256];
	size_t curarg;
	char *tempchr;
	char *tempchr2;
	static char data[513];
	char mydata[513];
	short length;

	curarg = strlen(c->buffer);

	args[0] = NULL;

	if (!strstr(c->buffer,"\r\n")) {
		if ((length = recv(c->s,&c->buffer[curarg],512 - curarg,0)) < 1) {
			irc_internal_disconnect(c,FE_DISCONNECT);
			firetalkerror = FE_DISCONNECT;
			return args;
		}
		c->buffer[curarg + length] = '\0';
	}
	if ((tempchr = strstr(c->buffer,"\r\n"))) {
		tempchr[0] = '\0';
		if (c->buffer[0] == ':')
			strcpy(data,&c->buffer[1]);
		else
			strcpy(data,c->buffer);
		strcpy(mydata,&tempchr[2]);
		strcpy(c->buffer,mydata);
	} else {
		if (strlen(c->buffer) >= 512) {
			irc_internal_disconnect(c,FE_PACKETSIZE);
			firetalkerror = FE_PACKETSIZE;
			return NULL;
		}
		firetalkerror = FE_SUCCESS;
		return args;
	}

	curarg = 0;
	tempchr = data;

	while ((curarg < 256) && ((tempchr2 = strchr(tempchr,' ')) != NULL)) {
		args[curarg++] = tempchr;
		tempchr2[0] = '\0';
		tempchr = tempchr2 + 1;
		if (tempchr[0] == ':') {
			if( strchr(tempchr,'\n') ) {
				strcpy(mydata,strchr(tempchr,'\n') + 1);
				strcat(mydata,"\r\n");
				strcpy(c->buffer,mydata);
			}
			tempchr = irc_irc_to_html(&tempchr[1],c->color_filter);
			break;
		}
	}
	args[curarg++] = tempchr;
	args[curarg] = NULL;
	return args;
}

static char *irc_get_nickname(const char * const hostmask) {
	static char data[512];
	char *tempchr = NULL;
	strcpy(data,hostmask);
	tempchr = strchr(data,'!');
	if (tempchr)
		tempchr[0] = '\0';
	return data;
}

static char *irc_get_address(const char * const hostmask) {
	static char data[512];
	char *tempchr = NULL;
	tempchr = strchr(hostmask,'@');
	if( tempchr && &tempchr[1] ) strcpy(data,&tempchr[1]);
	return data;
}

int irc_set_nickname(void * const c, const char * const nickname) {
	return irc_send_printf(c,"NICK %s",nickname);
}

int irc_set_password(void * const c, const char * const oldpass, const char * const newpass) {
	((struct s_irc_connection *)c)->passchange++;
	return irc_send_printf(c,"PRIVMSG NickServ :SET PASSWORD %s",newpass);
}

int irc_filter_colors(void * const c) {
	((struct s_irc_connection *)c)->color_filter = 1;
	return FE_SUCCESS;
}

int irc_allow_colors(void * const c) {
	((struct s_irc_connection *)c)->color_filter = 0;
	return FE_SUCCESS;
}

int irc_disconnect(void * const c) {
	irc_send_printf(c,"QUIT :User disconnected");
	return irc_internal_disconnect(c,FE_USERDISCONNECT);
}

void *irc_connect_server(const char * server, short port, const char * const nickname, const char * const password) {
	struct s_irc_connection *c;
	struct hostent *he;
	struct in_addr addr;
	struct sockaddr_in fulladdr;
	char ourserver[sizeof(IRC_SERVER)] = IRC_SERVER;
	int flag;
	char **args;

	if (!server) {
		server = ourserver;
		port = IRC_PORT;
	}
	
	c = malloc(sizeof(struct s_irc_connection));
	if (!c) {
		perror("malloc");
		exit(1);
	}
	memset(c, NULL, sizeof(struct s_irc_connection));
	c->nickname = NULL;
	c->lasttime = 0;
	c->isons = 0;
	c->color_filter = 1;
	c->whois_head = NULL;
	if( password ) {
		c->password = malloc(strlen(password) + 1);
		strcpy(c->password,password);
	}
	c->passchange = 0;
	c->usesilence = 1;

	if ( (c->s = socket(PF_INET, SOCK_STREAM, 0)) == -1) {
		irc_internal_disconnect(c,FE_SOCKET);
		firetalkerror = FE_SOCKET;
		return NULL;
	}

	if (! inet_aton(server,&addr)) {
		if (! (he = gethostbyname(server))) {
			irc_internal_disconnect(c,FE_RESOLV);
			firetalkerror = FE_RESOLV;
			return NULL;
		}
		memcpy(&addr,he->h_addr_list[0],sizeof(struct in_addr));
	}

	fulladdr.sin_family = AF_INET;
	fulladdr.sin_port = htons(port);
	memcpy(&fulladdr.sin_addr,&addr,sizeof(struct in_addr));

	if (connect(c->s,(const struct sockaddr *)&fulladdr,sizeof(struct sockaddr_in))) {
		irc_internal_disconnect(c,FE_CONNECT);
		firetalkerror = FE_CONNECT;
		return NULL;
	}

	if( password && strlen( password ) ) irc_send_printf(c,"PASS %s",password);
	irc_send_printf(c,"NICK %s",nickname);
	irc_send_printf(c,"USER %s %s %s :%s",nickname,server,server,getlogin());

	flag = 1;
	while (flag) {
		args = irc_recv_parse(c);
		if (!args[0])
			continue;
		/* zero argument items */
		if (!strcmp(args[0],"ERROR")) {
			irc_send_printf(c,"QUIT :error");
			irc_internal_disconnect(c,FE_CONNECT);
			firetalkerror = FE_CONNECT;
			return NULL;
		}

		if (!args[1])
			continue;
		/* one argument items */
		if (!strcmp(args[1],"ERROR")) {
			irc_send_printf(c,"QUIT :error");
			irc_internal_disconnect(c,FE_CONNECT);
			firetalkerror = FE_CONNECT;
			return NULL;
		}
		if (!strcmp(args[0],"PING"))
			irc_send_printf(c,"PONG %s",args[1]);
		else {
			switch (atoi(args[1])) {
				case 376:
					flag = 0;
					break;
				case 431:
				case 432:
				case 433:
				case 436:
				case 461:
					irc_send_printf(c,"QUIT :Invalid nickname");
					irc_internal_disconnect(c,FE_BADUSER);
					firetalkerror = FE_BADUSER;
					return NULL;
					break;
				case 465:
					irc_send_printf(c,"QUIT :banned");
					irc_internal_disconnect(c,FE_BLOCKED);
					firetalkerror = FE_BLOCKED;
					return NULL;
					break;
			}
		}
		
	}

	c->nickname = malloc(strlen(nickname) + 1);
	if (!c->nickname) {
		perror("malloc");
		exit(1);
	}
	strcpy(c->nickname,nickname);

	return c;
}

void irc_signon(void * const c) {
	struct s_irc_connection *conn;
	conn = c;
	firetalk_callback_doinit(c,conn->nickname);
}

int irc_got_data(void * const c) {
	struct s_irc_connection *conn;
	char **args;
	char *tempchr, *tempchr2, *tempchr3;
	struct s_firetalk_buddy *buddyiter;
	struct s_irc_whois *whoisiter, *whoisiter2;
	struct s_firetalk_handle *fchandle;
	int tempint,tempint2,tempint3;
	char tempbuf[512];
	time_t temptime;

	fchandle = firetalk_find_handle(c);

	conn = c;

	args = irc_recv_parse(conn);
	while (args[0]) {
		/* zero argument items */
		if (!strcmp(args[0],"PING")) {
			if (args[1])
				irc_send_printf(c,"PONG %s",args[1]);
			else
				irc_send_printf(c,"PONG");
		}
		if (args[1]) {
			tempint = atoi(args[1]);
			if (!strcmp(args[1],"QUIT")) {
				firetalk_callback_im_buddyonline(conn,irc_get_nickname(args[0]),0);
				irc_host_cache_remove(conn, args[0]);
				if (!irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])))
					irc_internal_disconnect(conn,FE_DISCONNECT);
				else
					firetalk_callback_chat_user_quit(conn,irc_get_nickname(args[0]));
			}
		}
		if (args[1] && args[2]) {
			/* two-arg commands */
			if (!strcmp(args[1],"JOIN")) {
				firetalk_callback_im_buddyonline(conn,irc_get_nickname(args[0]),1);
				irc_host_cache_add(conn, args[0]);
				if (!irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])))
					firetalk_callback_chat_joined(conn,&args[2][1]);
				else
					firetalk_callback_chat_user_joined(conn,&args[2][1],irc_get_nickname(args[0]),0);
			} else if (!strcmp(args[1],"PART")) {
				if (!irc_compare_nicks(conn->nickname,irc_get_nickname(args[0])))
					firetalk_callback_chat_left(conn,&args[2][1]);
				else
					firetalk_callback_chat_user_left(conn,&args[2][1],irc_get_nickname(args[0]));
			} else if (!strcmp(args[1],"NICK")) {
				irc_host_cache_change(conn,irc_get_nickname(args[0]),args[2]);
				if (!irc_compare_nicks(conn->nickname,irc_get_nickname(args[0]))) {
					firetalk_callback_newnick(conn,args[2]);
					tempchr = conn->nickname;
					conn->nickname = strdup(args[2]);
					if( tempchr ) free( tempchr );
				} else
					firetalk_callback_user_nickchanged(conn,irc_get_nickname(args[0]),args[2]);
			}
		}
		if (args[1] && args[2] && args[3]) {
			if (!strcmp(args[1],"PRIVMSG")) {
				irc_host_cache_add(conn, args[0]);
				/* scan for CTCP's */
				while ((tempchr = strchr(args[3],1))) {
					int correctctcp = 0;
					if ((tempchr2 = strchr(&tempchr[1],1))) {
						correctctcp = 1;
						tempchr2[0] = '\0';
					}
					/* we have a ctcp */
					if (!strcasecmp(&tempchr[1],"VERSION"))
						irc_send_printf(conn,"NOTICE %s :\001VERSION firetalk " LIBFIRETALK_VERSION "\001",irc_get_nickname(args[0]));
					else if (!strcasecmp(&tempchr[1],"SOURCE"))
						irc_send_printf(conn,"NOTICE %s :\001SOURCE " LIBFIRETALK_HOMEPAGE "\001",irc_get_nickname(args[0]));
					else if (!strcasecmp(&tempchr[1],"CLIENTINFO"))
						irc_send_printf(conn,"NOTICE %s :\001ACTION CLIENTINFO PING SOURCE TIME VERSION\001",irc_get_nickname(args[0]));
					else if (!strncasecmp(&tempchr[1],"PING ",5))
						irc_send_printf(conn,"NOTICE %s :\001PING %s\001",irc_get_nickname(args[0]),&tempchr[6]);
					else if (!strcasecmp(&tempchr[1],"TIME")) {
						time(&temptime);
						tempchr3 = ctime(&temptime);
						tempchr3[strlen(tempchr3)-1] = '\0';
						irc_send_printf(conn,"NOTICE %s :\001TIME %s\001",irc_get_nickname(args[0]),tempchr3);
					} else if (!strncasecmp(&tempchr[1],"ACTION ",7)) {
						if (args[2][0] == '#')
							firetalk_callback_chat_getaction(conn,&args[2][1],irc_get_nickname(args[0]),0,&tempchr[8]);
						else
							firetalk_callback_im_getaction(conn,irc_get_nickname(args[0]),0,&tempchr[8]);
					} else if (!strncasecmp(&tempchr[1],"DCC SEND ",9)) {
						char file[512];
						struct s_irc_dcc_connection *dcc_conn = (struct s_irc_dcc_connection *) malloc( sizeof( struct s_irc_dcc_connection ) );
						int accept = 0, resume = 0;
						memset( dcc_conn, NULL, sizeof( struct s_irc_dcc_connection ) );
						sscanf( args[3], "%*c%*s %*s %s %lu %hu %lu", file, &dcc_conn->host, &dcc_conn->port, &dcc_conn->size );
						tempchr3 = name_with_spaces( file );
						firetalk_callback_file_accept(conn,tempchr3,irc_get_nickname(args[0]),dcc_conn->size,&accept,&resume,&dcc_conn->path);
						free( tempchr3 );
						if( accept && strlen( dcc_conn->path ) ) {
							struct stat statbuf;
							dcc_conn->irc = conn;
							dcc_conn->filename = name_with_spaces( file );
							dcc_conn->user = strdup( irc_get_nickname(args[0]) );
							dcc_conn->position = 0;
							irc_dcc_conenction_add( conn, dcc_conn );
							if( resume && ! stat( dcc_conn->path, &statbuf ) && (u_long) statbuf.st_size <= dcc_conn->size ) {
								dcc_conn->position = (u_long) statbuf.st_size;
								tempchr3 = name_with_underscores( dcc_conn->filename );
								snprintf( tempbuf, 512, "%s %hu %lu", tempchr3, dcc_conn->port, dcc_conn->position );
								irc_send_printf(c,"PRIVMSG %s :\001DCC RESUME %s\001", irc_get_nickname(args[0]),tempbuf);
								free( tempchr3 );
							} else pthread_create( &dcc_conn->thread, NULL, irc_dcc_receive_file, dcc_conn );
						} else free( dcc_conn );
					} else if (!strncasecmp(&tempchr[1],"DCC RESUME ",11)) {
						struct s_irc_dcc_connection *dcc_conn = NULL;
						u_short port;
						u_long position;
						sscanf( args[3], "%*c%*s %*s %*s %hu %lu", &port, &position );
						if( ( dcc_conn = irc_dcc_conenction_find_by_port( conn, port ) ) ) {
							dcc_conn->position = position;
							tempchr3 = name_with_underscores( dcc_conn->filename );
							snprintf( tempbuf, 512, "%s %hu %lu",tempchr3,dcc_conn->port,dcc_conn->position);
							irc_send_printf(c,"PRIVMSG %s :\001DCC ACCEPT %s\001", irc_get_nickname(args[0]),tempbuf);
							free( tempchr3 );
						}
					} else if (!strncasecmp(&tempchr[1],"DCC ACCEPT ",11)) {
						struct s_irc_dcc_connection *dcc_conn = NULL;
						u_short port;
						u_long position;
						sscanf( args[3], "%*c%*s %*s %*s %hu %lu", &port, &position );
						if( ( dcc_conn = irc_dcc_conenction_find_by_port( conn, port ) ) && dcc_conn->position == position )
							pthread_create( &dcc_conn->thread, NULL, irc_dcc_receive_file, dcc_conn );
					} else if (!strncasecmp(&tempchr[1],"DCC CHAT ",9)) {
						irc_im_send_message( conn, irc_get_nickname(args[0]), "This user's IRC software does not support DCC Chat at this time.", TRUE );
					}
					if (correctctcp)
						strcpy(tempchr,&tempchr2[1]);
				}
				if (args[3][0]) {
					if (args[2][0] == '#')
						firetalk_callback_chat_getmessage(conn,&args[2][1],irc_get_nickname(args[0]),0,args[3]);
					else
						firetalk_callback_im_getmessage(conn,irc_get_nickname(args[0]),0,args[3]);
				}
			} else if (!strcmp(args[1],"NOTICE")) {
				irc_host_cache_add(conn, args[0]);
				if (!strcasecmp(irc_get_nickname(args[0]),"NickServ")) {
					if (conn->password && strstr(args[3],"IDENTIFY")) {
						conn->identified = 0;
					/* nickserv seems to be asking us to identify ourselves, and we have a password */
						if( strlen( conn->password ) )
							irc_send_printf(conn,"PRIVMSG NickServ :IDENTIFY %s",conn->password);
					}
					if (strstr(args[3],"Password changed") && conn->passchange) {
						/* successful change */
						conn->passchange--;
						firetalk_callback_passchanged(conn);
					}
					if (strstr(args[3],"authentication required") && conn->passchange) {
						/* didn't log in with the right password initially, not happening */
						conn->identified = 0;
						conn->passchange--;
						firetalk_callback_error(conn,FE_NOCHANGEPASS,NULL);
					}
					if (strstr(args[3],"isn't registered") && conn->passchange) {
						/* nick not registered, fail */
						conn->passchange--;
						firetalk_callback_error(conn,FE_NOCHANGEPASS,NULL);
					}
					if (strstr(args[3],"Password accepted") != NULL) {
						/* we're recognized */
						conn->identified = 1;
					}
				}
				if (args[2][0] == '#')
					firetalk_callback_chat_getmessage(conn,&args[2][1],irc_get_nickname(args[0]),1,args[3]);
				else
					firetalk_callback_im_getmessage(conn,irc_get_nickname(args[0]),1,args[3]);
			} else if (!strcmp(args[1],"TOPIC")) {
				irc_host_cache_add(conn, args[0]);
				firetalk_callback_chat_gottopic(conn,&args[2][1],args[3],irc_get_nickname(args[0]));
			} else if (!strcmp(args[1],"INVITE")) {
				irc_host_cache_add(conn, args[0]);
				firetalk_callback_chat_invited(conn,&args[3][1],irc_get_nickname(args[0]),NULL);
			} else if (!strcmp(args[1],"KICK")) {
				irc_host_cache_add(conn, args[0]);
				if (!irc_compare_nicks(conn->nickname,irc_get_nickname(args[3])))
					firetalk_callback_chat_kicked(conn,&args[2][1],irc_get_nickname(args[0]),args[4]);
				else {
					tempchr = strdup(irc_get_nickname(args[0]));
					firetalk_callback_chat_user_kicked(conn,&args[2][1],irc_get_nickname(args[3]),tempchr,args[4]);
					free(tempchr);
				}
			} else {
				switch (tempint) {
					case 301: /* RPL_AWAY */
						buddyiter = fchandle->buddy_head;
						while (buddyiter) {
							if (!irc_compare_nicks(args[3],buddyiter->nickname)) {
								buddyiter->tempint2 = 1;
								firetalk_callback_im_buddyaway(conn,args[3],1);
							}
							buddyiter = buddyiter->next;
						}
						break;
					case 302: /* RPL_USERHOST */ {
						tempchr = strchr(args[3],'*');
						if( tempchr ) {
							tempchr[0] = '\0';
							if( tempchr ) tempchr2 = &tempchr[3];
						} else {
							tempchr = strchr(args[3],'=');
							if( tempchr ) tempchr[0] = '\0';
							if( tempchr ) tempchr2 = &tempchr[2];
						}
						snprintf( tempbuf, 512, "%s!%s", args[3], tempchr2 );
						irc_host_cache_add(conn, tempbuf);
						break;
					} case 303: /* RPL_ISON */
						tempchr = args[3];
						while (tempchr && tempchr[0]) {
							tempchr2 = strchr(tempchr,' ');
							if (tempchr2)
								tempchr2[0] = '\0';
							buddyiter = fchandle->buddy_head;
							while (buddyiter) {
								if (!irc_compare_nicks(tempchr,buddyiter->nickname))
									buddyiter->tempint = 1;
								buddyiter = buddyiter->next;
							}
							if (tempchr2)
								tempchr = tempchr2 + 1;
							else
								tempchr = NULL;
						}
						if (--conn->isons <= 0) {
							conn->isons = 0;
							/* done, send the appropriate data */
							buddyiter = fchandle->buddy_head;
							while (buddyiter) {
								firetalk_callback_im_buddyonline(conn,buddyiter->nickname,buddyiter->tempint);
								if (buddyiter->tempint) {
									buddyiter->tempint2 = 0; /* away */
									irc_send_printf(conn,"WHOIS %s",buddyiter->nickname);
								}
								buddyiter = buddyiter->next;
							}
						}
						break;
					case 313: /* RPL_WHOISOPER */
						whoisiter = conn->whois_head;
						while (whoisiter) {
							if (!irc_compare_nicks(args[3],whoisiter->nickname))
								whoisiter->flags |= FF_ADMIN;
							whoisiter = whoisiter->next;
						}
						break;
					case 318: /* RPL_ENDOFWHOIS */
						whoisiter = conn->whois_head;
						whoisiter2 = NULL;
						while (whoisiter) {
							if (!irc_compare_nicks(args[3],whoisiter->nickname)) {
								/* manual whois */
								firetalk_callback_gotinfo(conn,whoisiter->nickname,whoisiter->info,0,whoisiter->idle,whoisiter->flags);
								free(whoisiter->nickname);
								if (whoisiter->info)
									free(whoisiter->info);
								if (whoisiter2)
									whoisiter2->next = whoisiter->next;
								else
									conn->whois_head = whoisiter->next;
								free(whoisiter);
								break;
							}
							whoisiter2 = whoisiter;
							whoisiter = whoisiter->next;
						}
						buddyiter = fchandle->buddy_head;
						while (buddyiter) {
							if (!irc_compare_nicks(args[3],buddyiter->nickname)) {
								if (!buddyiter->tempint2)
									firetalk_callback_im_buddyaway(conn,buddyiter->nickname,0);
							}
							buddyiter = buddyiter->next;
						}
						break;
					case 401: /* ERR_NOSUCHNICK */
					case 441: /* ERR_USERNOTINCHANNEL */
					case 443: /* ERR_USERONCHANNEL */
						firetalk_callback_im_buddyonline(conn,args[3],0);
						firetalk_callback_error(conn,FE_BADUSER,args[3]);
						if (!strcasecmp(args[3],"NickServ") && conn->passchange)
							conn->passchange--;
						whoisiter = conn->whois_head;
						whoisiter2 = NULL;
						while (whoisiter) {
							if (!irc_compare_nicks(args[3],whoisiter->nickname)) {
								free(whoisiter->nickname);
								if (whoisiter->info)
									free(whoisiter->info);
								if (whoisiter2)
									whoisiter2->next = whoisiter->next;
								else
									conn->whois_head = whoisiter->next;
								free(whoisiter);
								break;
							}
							whoisiter2 = whoisiter;
							whoisiter = whoisiter->next;
						}
						break;
					case 403: /* ERR_NOSUCHCHANNEL */
					case 442: /* ERR_NOTONCHANNEL */
						firetalk_callback_error(conn,FE_BADROOM,&args[3][1]);
						break;
					case 404: /* ERR_CANNOTSENDTOCHAN */
					case 405: /* ERR_TOOMANYCHANNELS */
					case 471: /* ERR_CHANNELISFULL */
					case 473: /* ERR_INVITEONLYCHAN */
					case 474: /* ERR_BANNEDFROMCHAN */
					case 475: /* ERR_BADCHANNELKEY */
						firetalk_callback_error(conn,FE_ROOMUNAVAILABLE,&args[3][1]);
						break;
					case 412: /* ERR_NOTEXTTOSEND */
						firetalk_callback_error(conn,FE_BADMESSAGE,NULL);
						break;
					case 421: /* ERR_UNKNOWNCOMMAND */
						if (!strcmp(args[3],"SILENCE"))
							conn->usesilence = 0;
						break;
					case 431: /* ERR_NONICKNAMEGIVEN */
					case 432: /* ERR_ERRONEUSNICKNAME */
					case 433: /* ERR_NICKNAMEINUSE */
					case 436: /* ERR_NICKCOLLISION */
						firetalk_callback_error(conn,FE_BADUSER,NULL);
						break;
					case 482: /* ERR_CHANOPRIVSNEEDED */
						firetalk_callback_error(conn,FE_NOPERMS,&args[3][1]);
						break;
				}
			}
		}
		if (args[1] && args[2] && args[3] && args[4]) {
			if (!strcmp(args[1],"MODE")) {
				tempint = 0;
				tempint2 = 4;
				tempint3 = 0;
				while (args[tempint2] && args[3][tempint]) {
					switch (args[3][tempint++]) {
						case '+':
							tempint3 = 1;
							break;
						case '-':
							tempint3 = -1;
							break;
						case 'o':
							if (tempint3 == 1)
								firetalk_callback_chat_user_opped(conn,&args[2][1],args[tempint2++],irc_get_nickname(args[0]));
							else if (tempint3 == -1)
								firetalk_callback_chat_user_deopped(conn,&args[2][1],args[tempint2++],irc_get_nickname(args[0]));
							break;
						case 'v':
							if (tempint3 == 1)
								firetalk_callback_chat_user_voiced(conn,&args[2][1],args[tempint2++],irc_get_nickname(args[0]));
							else if (tempint3 == -1)
								firetalk_callback_chat_user_devoiced(conn,&args[2][1],args[tempint2++],irc_get_nickname(args[0]));
							break;
						default:
							tempint2++;
							break;
					}
				}
			} else {
				switch (tempint) {
					case 317: /* RPL_WHOISIDLE */
						whoisiter = conn->whois_head;
						while (whoisiter) {
							if (!irc_compare_nicks(args[3],whoisiter->nickname))
								whoisiter->idle = atol(args[4])/60;
							whoisiter = whoisiter->next;
						}
						buddyiter = fchandle->buddy_head;
						while (buddyiter) {
							if (!irc_compare_nicks(args[3],buddyiter->nickname))
								firetalk_callback_idleinfo(conn,args[3],atol(args[4]) / 60);
							buddyiter = buddyiter->next;
						}
						break;
					case 332: /* RPL_TOPIC */
						firetalk_callback_chat_gottopic(conn,&args[3][1],args[4],NULL);
						break;
				}
			}
		}
		if (args[1] && args[2] && args[3] && args[4] && args[5]) {
			if (tempint == 353) {
				tempchr2 = args[5];
				while ((tempchr = strchr(tempchr2,' '))) {
					tempchr[0] = '\0';
					if (tempchr2[0] == '@' || tempchr2[0] == '+') {
						firetalk_callback_chat_user_joined(conn,&args[4][1],&tempchr2[1],1);
						firetalk_callback_im_buddyonline(conn,&tempchr2[1],1);
						irc_send_printf(c,"USERHOST %s",&tempchr2[1]);
						if (tempchr2[0] == '@')
							firetalk_callback_chat_user_opped(conn,&args[4][1],&tempchr2[1],NULL);
						else if (tempchr2[0] == '+')
							firetalk_callback_chat_user_voiced(conn,&args[4][1],&tempchr2[1],NULL);
					} else {
						firetalk_callback_chat_user_joined(conn,&args[4][1],tempchr2,1);
						firetalk_callback_im_buddyonline(conn,tempchr2,1);
						irc_send_printf(c,"USERHOST %s",tempchr2);
					}
					tempchr2 = tempchr + 1;
				}
			}
		}
		if (args[1] && args[2] && args[3] && args[4] && args[5] && args[6] && args[7]) {
			switch (tempint) {
				case 311: /* RPL_WHOISUSER */
					whoisiter = conn->whois_head;
					while (whoisiter) {
						if (!irc_compare_nicks(args[3],whoisiter->nickname)) {
							if (whoisiter->info)
								free(whoisiter->info);
							snprintf(tempbuf,512,"%s@%s: %s",args[4],args[5],args[7]);
							whoisiter->info = malloc(strlen(tempbuf)+1);
							strcpy(whoisiter->info,tempbuf);
						}
						whoisiter = whoisiter->next;
					}
				break;
			}
		}

		if (strstr(conn->buffer,"\r\n"))
			args = irc_recv_parse(conn);
		else
			args[0] = NULL;
	}

/* this is here so we can clean up dcc threads from the main thread (only needed every so often) */
	irc_dcc_conenction_cleanup(conn);

	firetalkerror = FE_SUCCESS;
	return FE_SUCCESS;
}

int irc_compare_nicks(const char * const nick1, const char * const nick2) {
	return strcasecmp(nick1,nick2);
}

int irc_get_fd(void * const c) {
	return ((struct s_irc_connection *)c)->s;
}

int irc_chat_join(void * const c, const char * const room) {
	return irc_send_printf(c,"JOIN #%s",room);
}

int irc_chat_part(void * const c, const char * const room) {
	return irc_send_printf(c,"PART #%s",room);
}

int irc_chat_send_message(void * const c, const char * const room, const char * const message, const int auto_flag) {
	if (auto_flag)
		return irc_send_printf(c,"NOTICE #%s :%s",room,message);
	else
		return irc_send_printf(c,"PRIVMSG #%s :%s",room,message);
}

int irc_chat_send_action(void * const c, const char * const room, const char * const message, const int auto_flag) {
	if (auto_flag)
		return irc_send_printf(c,"NOTICE #%s :\001ACTION %s\001",room,message);
	else
		return irc_send_printf(c,"PRIVMSG #%s :\001ACTION %s\001",room,message);
}

int irc_chat_invite(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"INVITE %s #%s",who,room);
}

int irc_im_send_message(void * const c, const char * const dest, const char * const message, const int auto_flag) {
	if (auto_flag)
		return irc_send_printf(c,"NOTICE %s :%s",dest,message);
	else
		return irc_send_printf(c,"PRIVMSG %s :%s",dest,message);
}

int irc_im_send_action(void * const c, const char * const dest, const char * const message, const int auto_flag) {
	if (auto_flag)
		return irc_send_printf(c,"NOTICE %s :\001ACTION %s\001",dest,message);
	else
		return irc_send_printf(c,"PRIVMSG %s :\001ACTION %s\001",dest,message);
}

int irc_chat_set_topic(void * const c, const char * const room, const char * const topic) {
	return irc_send_printf(c,"TOPIC #%s :%s",room,topic);
}

int irc_chat_op(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"MODE #%s +o %s",room,who);
}

int irc_chat_deop(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"MODE #%s -o %s",room,who);
}

int irc_chat_voice(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"MODE #%s +v %s",room,who);
}

int irc_chat_devoice(void * const c, const char * const room, const char * const who) {
	return irc_send_printf(c,"MODE #%s -v %s",room,who);
}

int irc_chat_kick(void * const c, const char * const room, const char * const who, const char * const reason) {
	if (reason)
		return irc_send_printf(c,"KICK #%s %s :%s",room,who,reason);
	else
		return irc_send_printf(c,"KICK #%s %s",room,who);
}

int irc_im_add_buddy(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	conn = c;
	conn->isons++;
	return irc_send_printf(c,"ISON %s",nickname);
}

int irc_im_add_deny(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	conn = c;
	if (conn->usesilence)
		return irc_send_printf(c,"SILENCE +%s!*@*",nickname);
	else
		return FE_SUCCESS;
}

int irc_im_remove_deny(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	conn = c;
	if (conn->usesilence)
		return irc_send_printf(c,"SILENCE -%s!*@*",nickname);
	else
		return FE_SUCCESS;
}

int irc_get_info(void * const c, const char * const nickname) {
	struct s_irc_connection *conn;
	struct s_irc_whois *whoistemp;
	conn = c;
	whoistemp = conn->whois_head;
	conn->whois_head = malloc(sizeof(struct s_irc_whois));
	memset( conn->whois_head, NULL, sizeof( struct s_irc_whois ) );
	conn->whois_head->nickname = malloc(strlen(nickname)+1);
	strcpy(conn->whois_head->nickname,nickname);
	conn->whois_head->flags = FF_NORMAL;
	conn->whois_head->idle = 0;
	conn->whois_head->info = NULL;
	conn->whois_head->next = whoistemp;
	return irc_send_printf(conn,"WHOIS %s",nickname);
}

int irc_set_away(void * const c, const char * const message) {
	if (message)
		return irc_send_printf(c,"AWAY :%s",message);
	else
		return irc_send_printf(c,"AWAY");
}

int irc_check_buddies(struct s_firetalk_handle * const c) {
	struct s_irc_connection *conn;
	struct s_firetalk_buddy *buddyiter;
	char obuf[505];

	obuf[0] = '\0';

	conn = c->handle;
	if (conn->lasttime > (time(NULL) - 40))
		return FE_IDLEFAST;

	if (conn->isons > 0)
		return FE_IDLEFAST;

	buddyiter = c->buddy_head;
	while (buddyiter) {
		if (strlen(obuf) + strlen(buddyiter->nickname) > 502) {
			obuf[strlen(obuf)-1] = '\0';
			irc_send_printf(conn,"ISON %s",obuf);
			conn->isons++;
			obuf[0] = '\0';
		}
		strcat(obuf,buddyiter->nickname);
		strcat(obuf," ");
		buddyiter->tempint = 0;
		buddyiter = buddyiter->next;
	}

	obuf[strlen(obuf)-1] = '\0';
	if (strlen(obuf)) {
		irc_send_printf(conn,"ISON %s",obuf);
		conn->isons++;
	}

	time(&conn->lasttime);
	return FE_SUCCESS;
}

/* DCC support added by Timothy Hatcher (11/28/2001) */

#ifdef USE_BCOPY
#define memmove(x,y,z)	bcopy((y), (x), (z))
#endif

int irc_dcc_setup_listen_port( const u_short ip_port ) {
	int sock = 0, tries = 0, opt = 1;
	static struct sockaddr_in server;

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if( sock < 0 ) {
		perror( "irc_dcc_setup_listen_port(): opening stream socket" );
		firetalkerror = FE_SOCKET;
		return -1;
	}

#ifdef SO_REUSEADDR
	if( setsockopt( sock, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof( opt ) ) < 0 ) {
		perror( "irc_dcc_setup_listen_port(): setsockopt SO_REUSEADDR" );
		firetalkerror = FE_SOCKET;
		return -1;
	}
#endif

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	server.sin_port = htons( ip_port );

	for( tries = 0; tries < 10; tries++ ) {
		pthread_testcancel();
		if( bind( sock, (struct sockaddr *) &server, sizeof( server ) ) ) {
			if( tries >= 9 ) {
				perror( "irc_dcc_setup_listen_port(): binding stream socket" );
				return -1;
			}
			perror( "irc_dcc_setup_listen_port(): binding stream socket. retry in 20 seconds" );
			firetalkerror = FE_SOCKET;
			pthread_testcancel();
			sleep( 20 );
		} else break;
	}
	listen( sock, 64 );
	return sock;
}

int irc_dcc_setup_connect_port( const u_long host, const u_short port ) {
	int sock = 0;
	static struct sockaddr_in server;

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if( sock < 0 ) {
		perror( "irc_dcc_setup_connect_port(): opening stream socket" );
		firetalkerror = FE_SOCKET;
		return -1;
	}

	server.sin_family = AF_INET;
	server.sin_addr.s_addr = ntohl( host );
	server.sin_port = htons( port );

	pthread_testcancel();
	if( connect( sock, (struct sockaddr *) &server, sizeof( server ) ) < 0 ) {
		perror( "irc_dcc_setup_connect_port(): connecting remote socket" );
		firetalkerror = FE_SOCKET;
		return -1;
	}

	return sock;
}

#define timeval_combine(tv) (tv.tv_sec + ((double)tv.tv_usec/1000000.))

#ifndef basename
char *basename( const char *name ) {
	const char *base = name;
	while( *name ) if( *name++ == '/' ) base = name;
	return (char *) base;
}
#endif

char *name_with_underscores( const char *name ) {
	char *ret = strdup( name );
	unsigned i;
	for( i = 0; ret[i]; i++ ) if( ret[i] == ' ' ) ret[i] = '_';
	return ret;
}

char *name_with_spaces( const char *name ) {
	char *ret = strdup( name );
	unsigned i;
	for( i = 0; ret[i]; i++ ) if( ret[i] == '_' ) ret[i] = ' ';
	return ret;
}

unsigned long primaryaddress( const char * const host ) {   
	struct hostent *hp = NULL;
	unsigned long addr = 0;

	addr = inet_addr( host );
	if( ! addr || addr == INADDR_NONE ) hp = gethostbyname( host );
	if( hp ) memmove( &addr, hp->h_addr_list[0], 4 );

	return addr;
}

int irc_file_send(void * const c, const char * const who, const char * const filepath) {
	static u_short rport = 1024;
	struct s_irc_connection *conn = c;
	const struct s_irc_host_cache *cache = NULL;
	struct s_irc_dcc_connection *dcc_conn = NULL;
	if( ! ( cache = irc_host_cache_find(c,who) ) ) {
		irc_send_printf(c,"USERHOST %s",who);
		firetalkerror = FE_RESOLV;
		return FE_RESOLV;
	}
	if( ! cache->host ) {
		if( ! ( ((struct s_irc_host_cache *) cache)->host = primaryaddress( cache->address ) ) ) {
			firetalkerror = FE_RESOLV;
			return FE_RESOLV;
		}
	}
	dcc_conn = (struct s_irc_dcc_connection *) malloc( sizeof( struct s_irc_dcc_connection ) );
	memset( dcc_conn, NULL, sizeof( struct s_irc_dcc_connection ) );
	dcc_conn->irc = conn;
	dcc_conn->filename = strdup( basename( filepath ) );
	dcc_conn->path = strdup( filepath );
	dcc_conn->user = strdup( who );
	dcc_conn->host = cache->host;
	dcc_conn->port = rport++;
	dcc_conn->ifd = -1;
	dcc_conn->ofd = -1;
	dcc_conn->socket = -1;
	irc_dcc_conenction_add( conn, dcc_conn );
	pthread_create( &dcc_conn->thread, NULL, irc_dcc_send_file, dcc_conn );
	return FE_SUCCESS;
}

int irc_file_cancel(void * const c, const char * const who, const char * const filename) {
	struct s_irc_dcc_connection *dcc = NULL;

	if( ! ( dcc = irc_dcc_conenction_find_by_names( c, who, filename ) ) ) return FE_SUCCESS;

	irc_dcc_conenction_remove( c, dcc );

	return FE_SUCCESS;
}

/*
 * irc_dcc_send_file( void *c )
 * listens to connections to port, and when connection established
 * sends ifile to that socket
 */
void *irc_dcc_send_file( void *c ) {
	struct s_irc_dcc_connection *dcc = c;
	ssize_t len = 0;
	u_long bytessent = 0, prev = 0;
	char buf[1024 * 8], args[512], *filename = NULL;
	char namebuf[31];
	struct stat statbuf;
	struct hostent *hp;
	struct timeval start, end;
	double slice_time = 0.;

	dcc->finished = 0;

	if( ( dcc->ifd = open( dcc->path, O_RDONLY ) ) < 0 ) {
		fprintf( stderr, "ERROR opening file: %s\n", dcc->path );
		firetalkerror = FE_NOPERMS;
		goto exit;
	}

	gethostname( namebuf, sizeof( namebuf ) );
	fstat( dcc->ifd, &statbuf );
	dcc->size = (u_long) statbuf.st_size;
	hp = gethostbyname( namebuf );

	if( ( dcc->socket = irc_dcc_setup_listen_port( dcc->port ) ) < 0 ) {
		firetalkerror = FE_SOCKET;
		goto exit;
	}

	if( hp ) {
		filename = name_with_underscores( dcc->filename );
		snprintf( args, 512, "%s %lu %hu %lu", filename, (unsigned long) ntohl( ( (struct in_addr *)( hp->h_addr_list)[0] )->s_addr ), dcc->port, dcc->size );
		irc_send_printf( dcc->irc, "PRIVMSG %s :\001DCC SEND %s\001", dcc->user, args );
		free( filename );
	} else {
		firetalkerror = FE_RESOLV;
		goto exit;
	}

	pthread_testcancel();
	dcc->ofd = accept( dcc->socket, NULL, NULL );
	pthread_testcancel();

	lseek( dcc->ifd, dcc->position, SEEK_SET );
	bytessent = dcc->position;
	prev = bytessent;

	filename = name_with_spaces( dcc->filename );
	gettimeofday( &start, NULL );
	slice_time = timeval_combine( start );
	firetalk_callback_file_progress( dcc->irc, filename, dcc->user, dcc->position, dcc->size, 0, 0., FALSE );
	while( ( len = read( dcc->ifd, buf, sizeof( buf ) ) ) > 0 ) {
		write( dcc->ofd, buf, len );
		bytessent += len;
		if( dcc->finished ) goto exit;
		if( bytessent - prev > dcc->size / 200 ) {
			gettimeofday( &end, NULL );
			slice_time = timeval_combine( end ) - slice_time;
			firetalk_callback_file_progress( dcc->irc, filename, dcc->user, bytessent, dcc->size, bytessent - prev, slice_time, FALSE );
			gettimeofday( &start, NULL );
			slice_time = timeval_combine( start );
			prev = bytessent;
		}
		while( ( len = read( dcc->ofd, buf, sizeof( u_long ) ) ) &&
			ntohl( *(u_long *) buf ) != bytessent ) pthread_testcancel();
		if( len <= 0 ) break;
	}
	gettimeofday( &end, NULL );
	slice_time = timeval_combine( end ) - slice_time;
exit:
	firetalk_callback_file_progress( dcc->irc, filename, dcc->user, bytessent, dcc->size, bytessent - prev, slice_time, TRUE );
	free( filename );
	close( dcc->ofd );
	close( dcc->ifd );
	dcc->ifd = -1;
	dcc->ofd = -1;
	dcc->finished = 1;
	return NULL;
}

/*
 * irc_dcc_receive_file( void *c )
 * connects to (c->host,c->port) and reads everything send from there
 * for every packet received gives back how much actually got
 * puts everything in c->saveto file
 */
void *irc_dcc_receive_file( void *c ) {
	struct s_irc_dcc_connection *dcc = c;
	char buf[1024 * 8], *filename = NULL;
	u_long netsize = 0, bytesreceived = 0, prev = 0;
	ssize_t len = 0, toread = 0;
	double slice_time = 0.;
	struct timeval start, end;

	dcc->finished = 0;

	if( ( dcc->ofd = open( dcc->path, O_WRONLY|O_CREAT|( dcc->position ? 0 : O_TRUNC ), 0644 ) ) < 0 ) {
		fprintf( stderr, "ERROR opening file: %s\n", dcc->path );
		firetalkerror = FE_NOPERMS;
		goto exit;
	}

	lseek( dcc->ofd, dcc->position, SEEK_SET );
	bytesreceived = dcc->position;
	prev = bytesreceived;

	filename = name_with_spaces( dcc->filename );
	dcc->ifd = irc_dcc_setup_connect_port( dcc->host, dcc->port );
	if( dcc->ifd < 0 ) goto exit;
	if( ( toread = sizeof( buf ) ) > dcc->size ) toread = dcc->size;
	gettimeofday( &start, NULL );
	slice_time = timeval_combine( start );
	firetalk_callback_file_progress( dcc->irc, filename, dcc->user, bytesreceived, dcc->size, 0, 0., FALSE );
	while( bytesreceived < dcc->size && ( len = read( dcc->ifd, buf, toread ) ) > 0 ) {
		write( dcc->ofd, buf, len );
		bytesreceived += len;
		if( dcc->finished ) goto exit;
		usleep( 1000 );
		netsize = htonl( bytesreceived );
		lseek( dcc->ifd, 0, SEEK_END );
		write( dcc->ifd, &netsize, sizeof( u_long ) );
		lseek( dcc->ifd, 0, SEEK_END );
		if( toread > dcc->size - bytesreceived ) toread = dcc->size - bytesreceived;
		if( bytesreceived - prev > dcc->size / 200 ) {
			gettimeofday( &end, NULL );
			slice_time = timeval_combine( end ) - slice_time;
			firetalk_callback_file_progress( dcc->irc, filename, dcc->user, bytesreceived, dcc->size, bytesreceived - prev, slice_time, FALSE );
			gettimeofday( &start, NULL );
			slice_time = timeval_combine( start );
			prev = bytesreceived;
		}
		pthread_testcancel();
	}
	gettimeofday( &end, NULL );
	slice_time = timeval_combine( end ) - slice_time;
exit:
	firetalk_callback_file_progress( dcc->irc, filename, dcc->user, bytesreceived, dcc->size, bytesreceived - prev, slice_time, TRUE );
	free( filename );
	close( dcc->ifd );
	close( dcc->ofd );
	dcc->ifd = -1;
	dcc->ofd = -1;
	dcc->finished = 1;
	printf( "finish\n" );
	return NULL;
}

void irc_dcc_conenction_add(struct s_irc_connection * const conn,struct s_irc_dcc_connection * const dcc) {
	struct s_irc_dcc_connection *cur = NULL;

	if( ! dcc ) return;
	if( irc_dcc_conenction_find_by_port( conn, dcc->port ) ) return;

	for( cur = conn->dcc_active; cur; cur = cur->next )
		if( ! cur->next && cur ) break;

	if( cur ) cur->next = dcc;
	else if( ! conn->dcc_active ) conn->dcc_active = dcc;
	dcc->back = cur;
	dcc->next = NULL;
}

void irc_dcc_conenction_remove(struct s_irc_connection * const conn,struct s_irc_dcc_connection * const dcc) {
	struct s_irc_dcc_connection *back = NULL, *next = NULL;

	if( ! dcc ) return;
	back = dcc->back;
	next = dcc->next;

	if( conn->dcc_active == dcc ) conn->dcc_active = next;
	if( back ) back->next = next;
	if( next ) next->back = back;

	pthread_cancel( dcc->thread );

	dcc->finished = 1;
	if( dcc->ifd > 0 ) close( dcc->ifd );
	if( dcc->ofd > 0 ) close( dcc->ofd );
	if( dcc->socket > 0 ) close( dcc->socket );
	dcc->ifd = dcc->ofd = dcc->socket = -1;

	pthread_join( dcc->thread, NULL );

	free( dcc->filename );
	free( dcc->path );
	free( dcc->user );
	free( (struct s_irc_dcc_connection *) dcc );
}

s_irc_dcc_connection * const irc_dcc_conenction_find_by_port(struct s_irc_connection * const conn,const u_short port) {
	struct s_irc_dcc_connection *cur = NULL;
	if( ! port ) return NULL;
	irc_dcc_conenction_cleanup( conn );
	for( cur = conn->dcc_active; cur; cur = cur->next )
		if( cur->port == port )
			return cur;
	return NULL;
}

s_irc_dcc_connection * const irc_dcc_conenction_find_by_names(struct s_irc_connection * const conn,const char * const who,const char * const file) {
	struct s_irc_dcc_connection *cur = NULL;
	char *temp = NULL;
	if( ! who || ! file ) return NULL;
	irc_dcc_conenction_cleanup( conn );
	temp = name_with_underscores( file );
	for( cur = conn->dcc_active; cur; cur = cur->next )
		if( ! strcmp( cur->user, who ) && ( ! strcmp( cur->path, file ) || ! strcmp( cur->filename, file ) || ! strcmp( cur->filename, temp ) ) )
			break;
	free( temp );
	return cur;
}

void irc_dcc_conenction_cleanup(struct s_irc_connection * const conn) {
	struct s_irc_dcc_connection *cur = NULL, *tmp;
	if( ! conn ) return;
	for( cur = conn->dcc_active; cur; ) {
		if( cur->finished ) {
			tmp = cur->next;
			irc_dcc_conenction_remove( conn, cur );
			cur = tmp;
		} else cur = cur->next;
	}
}

void irc_host_cache_add(struct s_irc_connection * const conn,const char * const hostmask) {
	struct s_irc_host_cache *new = NULL, *cur = NULL;

	if( irc_host_cache_find( conn, irc_get_nickname( hostmask ) ) ) return;
	if( ! strlen( irc_get_address( hostmask ) ) ) return;

	if( ! ( new = (struct s_irc_host_cache *) malloc( sizeof( struct s_irc_host_cache ) ) ) ) {
		exit(1);
	}
	memset( new, NULL, sizeof( struct s_irc_host_cache ) );
	new->nickname = strdup( irc_get_nickname( hostmask ) );
	new->address = strdup( irc_get_address( hostmask ) );
//	new->host = primaryaddress( new->address );

	for( cur = conn->host_cache; cur; cur = cur->next )
		if( ! cur->next && cur ) break;

	if( cur ) cur->next = new;
	else if( ! conn->host_cache ) conn->host_cache = new;
	new->back = cur;
	new->next = NULL;
}

void irc_host_cache_remove(struct s_irc_connection * const conn,const char * const hostmask) {
	struct s_irc_host_cache *back = NULL, *next = NULL;
	const struct s_irc_host_cache *find = NULL;

	if( ! ( find = irc_host_cache_find( conn, irc_get_nickname( hostmask ) ) ) ) return;
	back = find->back;
	next = find->next;

	if( conn->host_cache == find ) conn->host_cache = next;
	if( back ) back->next = next;
	if( next ) next->back = back;

	free( find->nickname );
	free( find->address );
	free( (struct s_irc_host_cache *) find );
	find = NULL;
}

void irc_host_cache_change(struct s_irc_connection * const conn,const char * const old,const char * const new) {
	struct s_irc_host_cache *find = NULL;
	if( ! ( find = (struct s_irc_host_cache *) irc_host_cache_find( conn, old ) ) ) return;
	free( find->nickname );
	find->nickname = strdup( new );
}

struct s_irc_host_cache * const irc_host_cache_find(struct s_irc_connection * const conn,const char * const user) {
	struct s_irc_host_cache *cur = NULL;
	if( ! conn || ! user ) return NULL;
	for( cur = conn->host_cache; cur; cur = cur->next )
		if( ! strcmp( cur->nickname, user ) )
			return cur;
	return NULL;
}

void irc_host_cache_free(struct s_irc_connection * const conn) {
	struct s_irc_host_cache *cur = NULL, *next = NULL;

	if( ! conn ) return;
	cur = conn->host_cache;
	while( cur ) {
		next = cur->next;
		free( cur->nickname );
		free( cur->address );
		free( cur );
		cur = NULL;
		cur = next;
	}
	conn->host_cache = NULL;
}
